"""This module implements the Ellipse shape entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# Conflict with Isort
# pylint: disable=wrong-import-order

import datetime
import math
from typing import List, Optional, Tuple

import numpy as np
from scipy import optimize, special
from shapely.geometry import Polygon as shapely_polygon

from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import Shape, ShapeType
from otx.api.utils.time_utils import now

# pylint: disable=invalid-name


class Ellipse(Shape):
    """Ellipse represents an ellipse that is encapsulated by a Rectangle.

    - x1 and y1 represent the top-left coordinate of the encapsulating rectangle
    - x2 and y2 representing the bottom-right coordinate of the encapsulating rectangle

    Args:
        x1: left x coordinate of encapsulating rectangle
        y1: top y coordinate of encapsulating rectangle
        x2: right x coordinate of encapsulating rectangle
        y2: bottom y coordinate of encapsulating rectangle
            modification_date: last modified date
    """

    # pylint: disable=too-many-arguments; Requires refactor
    def __init__(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        modification_date: Optional[datetime.datetime] = None,
    ):
        modification_date = now() if modification_date is None else modification_date
        super().__init__(
            shape_type=ShapeType.ELLIPSE,
            modification_date=modification_date,
        )

        for (x, y) in [(x1, y1), (x2, y2)]:
            self._validate_coordinates(x, y)

        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        if self.width <= 0 or self.height <= 0:
            raise ValueError(
                f"Invalid Ellipse with coordinates: x1={self.x1}, y1={self.y1}, x2={self.x2}," f" y2={self.y2}"
            )

    def __repr__(self):
        """Returns the representation of the Ellipse."""
        return f"Ellipse(x1={self.x1}, y1={self.y1}, x2={self.x2}, y2={self.y2})"

    def __eq__(self, other):
        """Returns True if Ellipse is equal to other."""
        if isinstance(other, Ellipse):
            return (
                self.x1 == other.x1
                and self.y1 == other.y1
                and self.x2 == other.x2
                and self.y2 == other.y2
                and self.modification_date == other.modification_date
            )
        return False

    def __hash__(self):
        """Returns the hash of the Ellipse."""
        return hash(str(self))

    @property
    def width(self) -> float:
        """Returns the width [x-axis] of the ellipse.

        Example:

        >>> e1 = Ellipse(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
        >>> e1.width
        0.5

        Returns:
            the width of the ellipse. (x-axis)
        """
        return self.x2 - self.x1

    @property
    def height(self) -> float:
        """Returns the height [y-axis] of the ellipse.

        Example:

        >>> e1 = Ellipse(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
        >>> e1.height
        0.5

        Returns:
            the height of the ellipse. (y-axis)
        """
        return self.y2 - self.y1

    @property
    def x_center(self) -> float:
        """Returns the x coordinate in the center of the ellipse."""
        return self.x1 + self.width / 2

    @property
    def y_center(self) -> float:
        """Returns the y coordinate in the center of the ellipse."""
        return self.y1 + self.height / 2

    @property
    def minor_axis(self) -> float:
        """Returns the minor axis of the ellipse.

        Example:

        >>> e1 = Ellipse(x1=0.5, x2=1.0, y1=0.0, y2=0.4)
        >>> e1.minor_axis
        0.2

        Returns:
            minor axis of ellipse.
        """
        if self.width > self.height:
            return self.height / 2
        return self.width / 2

    @property
    def major_axis(self) -> float:
        """Returns the major axis of the ellipse.

        Example:

        >>> e1 = Ellipse(x1=0.5, x2=1.0, y1=0.0, y2=0.4)
        >>> e1.major_axis
        0.25

        Returns:
            major axis of ellipse.
        """
        if self.height > self.width:
            return self.height / 2
        return self.width / 2

    def normalize_wrt_roi_shape(self, roi_shape: Rectangle) -> "Ellipse":
        """Transforms from the `roi` coordinate system to the normalized coordinate system.

        This function is the inverse of ``denormalize_wrt_roi_shape``.

        Example:
            Assume we have Ellipse `c1` which lives in the top-right quarter of a 2D space.
            The 2D space where `c1` lives in is an `roi` living in the top-left quarter of the normalized coordinate
            space. This function returns Ellipse `c1` expressed in the normalized coordinate space.

            >>> from otx.api.entities.annotation import Annotation
            >>> from otx.api.entities.shapes.rectangle import Rectangle
            >>> from otx.api.entities.shapes.ellipse import Ellipse
            >>> c1 = Ellipse(x1=0.5, y1=0.5, x2=0.6, y2=0.6)
            >>> roi = Rectangle(x1=0.0, x2=0.5, y1=0.0, y2=0.5)
            >>> normalized = c1.normalize_wrt_roi_shape(roi_shape)
            >>> normalized
            Ellipse(, x1=0.25, y1=0.25, x2=0.3, y2=0.3)

        Args:
            roi_shape: Region of Interest

        Returns:
            New polygon in the image coordinate system
        """

        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        return Ellipse(
            x1=self.x1 * roi_shape.width + roi_shape.x1,
            y1=self.y1 * roi_shape.height + roi_shape.y1,
            x2=self.x2 * roi_shape.width + roi_shape.x1,
            y2=self.y2 * roi_shape.height + roi_shape.y1,
        )

    def denormalize_wrt_roi_shape(self, roi_shape: Rectangle) -> "Ellipse":
        """Transforming shape from the normalized coordinate system to the `roi` coordinate system.

        This function is the inverse of ``normalize_wrt_roi_shape``

        Example:
            Assume we have Ellipse `c1` which lives in the top-right quarter of the normalized coordinate space.
            The `roi` is a rectangle living in the half right of the normalized coordinate space.
            This function returns Ellipse `c1` expressed in the coordinate space of `roi`. (should return top-half)

            Ellipse denormalized to a rectangle as ROI

            >>> from otx.api.entities.annotation import Annotation
            >>> from otx.api.entities.shapes.ellipse import Ellipse
            >>> c1 = Ellipse(x1=0.5, x2=1.0, y1=0.0, y2=0.5)  # An ellipse in the top right
            >>> roi = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=1.0)  # the half-right
            >>> normalized = c1.denormalize_wrt_roi_shape(roi_shape)  # should return top half
            >>> normalized
            Ellipse(, x1=0.0, y1=0.0, x2=1.0, y2=0.5)

        Args:
            roi_shape: Region of Interest

        Returns:
            New polygon in the ROI coordinate system
        """
        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        x1 = (self.x1 - roi_shape.x1) / roi_shape.width
        y1 = (self.y1 - roi_shape.y1) / roi_shape.height
        x2 = (self.x2 - roi_shape.x1) / roi_shape.width
        y2 = (self.y2 - roi_shape.y1) / roi_shape.height

        return Ellipse(x1=x1, y1=y1, x2=x2, y2=y2)

    # pylint: disable=no-member; PyLint cannot find scipy.special.ellipeinc()
    def get_evenly_distributed_ellipse_coordinates(self, number_of_coordinates: int = 50) -> List[Tuple[float, float]]:
        """Returns evenly distributed coordinates along the ellipse.

        Makes use of scipy.special.ellipeinc() which provides the numerical integral along the perimeter of the ellipse,
        and scipy.optimize.root() for solving the equal-arcs length equation for the angles.

        Args:
            number_of_coordinates: number of evenly distributed points
                to generate along the ellipsis line

        Returns:
            list of tuple's with coordinates along the ellipse line
        """
        angles = 2 * np.pi * np.arange(number_of_coordinates) / number_of_coordinates
        e = (1.0 - self.minor_axis**2.0 / self.major_axis**2.0) ** 0.5
        total_size = special.ellipeinc(2.0 * np.pi, e)
        arc_size = total_size / number_of_coordinates
        arcs = np.arange(number_of_coordinates) * arc_size
        res = optimize.root(lambda x: (special.ellipeinc(x, e) - arcs), angles)
        angles = res.x
        if self.width > self.height:
            x_points = list(self.major_axis * np.sin(angles))
            y_points = list(self.minor_axis * np.cos(angles))
        else:
            x_points = list(self.minor_axis * np.cos(angles))
            y_points = list(self.major_axis * np.sin(angles))
        coordinates = [
            (point_x + self.x_center, point_y + self.y_center) for point_x, point_y in zip(x_points, y_points)
        ]
        return coordinates

    def _as_shapely_polygon(self) -> shapely_polygon:
        coordinates = self.get_evenly_distributed_ellipse_coordinates()
        return shapely_polygon(coordinates)

    def get_area(self) -> float:
        """Computes the approximate area of the Ellipse.

        Area is a value between 0 and 1, computed as
        `pi * vertex * co-vertex`.

            >>> Ellipse(x1=0, y1=0, x2=0.8, y2=0.4).get_area()
            0.25132741228718347

        Returns:
            area of the shape
        """
        return math.pi * self.minor_axis * self.major_axis



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math
from dataclasses import dataclass, field

import torch
from torch.nn import functional as F

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class SoftMaxV0Attribute(Attribute):
    axis: int = field(default=1)


@OPS.register()
class SoftMaxV0(Operation[SoftMaxV0Attribute]):
    TYPE = "Softmax"
    VERSION = 0
    ATTRIBUTE_FACTORY = SoftMaxV0Attribute

    def forward(self, input):
        return F.softmax(input=input, dim=self.attrs.axis)


@dataclass
class SoftMaxV1Attribute(Attribute):
    axis: int = field(default=1)


@OPS.register()
class SoftMaxV1(Operation[SoftMaxV1Attribute]):
    TYPE = "Softmax"
    VERSION = 1
    ATTRIBUTE_FACTORY = SoftMaxV1Attribute

    def forward(self, input):
        return F.softmax(input=input, dim=self.attrs.axis)


@dataclass
class ReluV0Attribute(Attribute):
    pass


@OPS.register()
class ReluV0(Operation[ReluV0Attribute]):
    TYPE = "Relu"
    VERSION = 0
    ATTRIBUTE_FACTORY = ReluV0Attribute

    def forward(self, input):
        return F.relu(input)


@dataclass
class SwishV4Attribute(Attribute):
    pass


@OPS.register()
class SwishV4(Operation[SwishV4Attribute]):
    TYPE = "Swish"
    VERSION = 4
    ATTRIBUTE_FACTORY = SwishV4Attribute

    def forward(self, input, beta=1.0):
        return input * torch.sigmoid(input * beta)


@dataclass
class SigmoidV0Attribute(Attribute):
    pass


@OPS.register()
class SigmoidV0(Operation[SigmoidV0Attribute]):
    TYPE = "Sigmoid"
    VERSION = 0
    ATTRIBUTE_FACTORY = SigmoidV0Attribute

    def forward(self, input):
        return torch.sigmoid(input)


@dataclass
class ClampV0Attribute(Attribute):
    min: float
    max: float


@OPS.register()
class ClampV0(Operation[ClampV0Attribute]):
    TYPE = "Clamp"
    VERSION = 0
    ATTRIBUTE_FACTORY = ClampV0Attribute

    def forward(self, input):
        return input.clamp(min=self.attrs.min, max=self.attrs.max)


@dataclass
class PReluV0Attribute(Attribute):
    pass


@OPS.register()
class PReluV0(Operation[PReluV0Attribute]):
    TYPE = "PRelu"
    VERSION = 0
    ATTRIBUTE_FACTORY = PReluV0Attribute

    def forward(self, input, slope):
        return F.prelu(input=input, weight=slope)


@dataclass
class TanhV0Attribute(Attribute):
    pass


@OPS.register()
class TanhV0(Operation[TanhV0Attribute]):
    TYPE = "Tanh"
    VERSION = 0
    ATTRIBUTE_FACTORY = TanhV0Attribute

    def forward(self, input):
        return F.tanh(input)


@dataclass
class EluV0Attribute(Attribute):
    alpha: float


@OPS.register()
class EluV0(Operation[EluV0Attribute]):
    TYPE = "Elu"
    VERSION = 0
    ATTRIBUTE_FACTORY = EluV0Attribute

    def forward(self, input):
        return F.elu(input=input, alpha=self.attrs.alpha)


@dataclass
class SeluV0Attribute(Attribute):
    pass


@OPS.register()
class SeluV0(Operation[SeluV0Attribute]):
    TYPE = "Selu"
    VERSION = 0
    ATTRIBUTE_FACTORY = SeluV0Attribute

    def forward(self, input, alpha, lambda_):
        return lambda_ * F.elu(input=input, alpha=alpha)


@dataclass
class MishV4Attribute(Attribute):
    pass


@OPS.register()
class MishV4(Operation[MishV4Attribute]):
    TYPE = "Mish"
    VERSION = 4
    ATTRIBUTE_FACTORY = MishV4Attribute

    def forward(self, input):
        # NOTE: pytorch 1.8.2 does not have mish function
        #  return F.mish(input=input)
        return input * F.tanh(F.softplus(input))


@dataclass
class HSwishV4Attribute(Attribute):
    pass


@OPS.register()
class HSwishV4(Operation[HSwishV4Attribute]):
    TYPE = "HSwish"
    VERSION = 4
    ATTRIBUTE_FACTORY = HSwishV4Attribute

    def forward(self, input):
        return F.hardswish(input=input)


@dataclass
class HSigmoidV5Attribute(Attribute):
    pass


@OPS.register()
class HSigmoidV5(Operation[HSigmoidV5Attribute]):
    TYPE = "HSigmoid"
    VERSION = 5
    ATTRIBUTE_FACTORY = HSigmoidV5Attribute

    def forward(self, input):
        return F.hardsigmoid(input=input)


@dataclass
class ExpV0Attribute(Attribute):
    pass


@OPS.register()
class ExpV0(Operation[ExpV0Attribute]):
    TYPE = "Exp"
    VERSION = 0
    ATTRIBUTE_FACTORY = ExpV0Attribute

    def forward(self, input):
        return torch.exp(input)


@dataclass
class HardSigmoidV0Attribute(Attribute):
    pass


@OPS.register()
class HardSigmoidV0(Operation[HardSigmoidV0Attribute]):
    TYPE = "HardSigmoid"
    VERSION = 0
    ATTRIBUTE_FACTORY = HardSigmoidV0Attribute

    def forward(self, input, alpha, beta):
        return torch.maximum(
            torch.zeros_like(input),
            torch.minimum(torch.ones_like(input), input * alpha + beta),
        )


@dataclass
class GeluV7Attribute(Attribute):
    approximation_mode: str = field(default="ERF")

    def __post_init__(self):
        super().__post_init__()
        valid_approximation_mode = ["ERF", "tanh"]
        if self.approximation_mode not in valid_approximation_mode:
            raise ValueError(
                f"Invalid approximation_mode {self.approximation_mode}. "
                f"It must be one of {valid_approximation_mode}."
            )


@OPS.register()
class GeluV7(Operation[GeluV7Attribute]):
    TYPE = "Gelu"
    VERSION = 7
    ATTRIBUTE_FACTORY = GeluV7Attribute

    def forward(self, input):
        mode = self.attrs.approximation_mode
        if mode == "ERF":
            return F.gelu(input=input)
        elif mode == "tanh":
            return input * 0.5 * (1 + F.tanh(torch.sqrt(2 / torch.tensor(math.pi)) * (input + 0.044715 * input**3)))



"""Tests the methods in the NNCF task."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

from otx.algorithms.anomaly.tasks.nncf import NNCFTask
from otx.algorithms.anomaly.tasks.train import TrainingTask
from otx.api.entities.model import ModelEntity, ModelOptimizationType
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType


class TestNNCFTask:
    """Tests methods in the NNCF task."""

    def test_nncf(self, tmpdir, setup_task_environment):
        """Tests the NNCF optimize method."""
        root = str(tmpdir.mkdir("anomaly_nncf_test"))

        # Get task environment
        setup_task_environment = deepcopy(setup_task_environment)  # since fixture is mutable
        task_environment = setup_task_environment.task_environment
        output_model = setup_task_environment.output_model
        dataset = setup_task_environment.dataset

        train_task = TrainingTask(task_environment, output_path=root)
        train_task.train(dataset, output_model, TrainParameters())
        task_environment.model = output_model

        # create normal nncf task
        nncf_task = NNCFTask(task_environment, output_path=root)

        # optimize the model
        assert output_model.optimization_type == ModelOptimizationType.NONE
        optimized_model = ModelEntity(dataset, configuration=task_environment.get_model_configuration())
        nncf_task.optimize(OptimizationType.NNCF, dataset, optimized_model)
        assert optimized_model.optimization_type == ModelOptimizationType.NNCF

        # load the optimized model
        new_nncf_task = NNCFTask(task_environment, output_path=root)
        assert new_nncf_task.compression_ctrl is None
        new_nncf_task.model = new_nncf_task.load_model(optimized_model)
        assert new_nncf_task.compression_ctrl is not None

        # Export model
        new_nncf_task.export(ExportType.OPENVINO, optimized_model)
        optimized_model.get_data("openvino.bin")  # Should not raise an exception



"""Base EfficientNet."""

# model settings
model = dict(
    type="ImageClassifier",
    backbone=dict(type="otx.OTXEfficientNet", pretrained=True, version="b0"),
    neck=dict(type="GlobalAveragePooling"),
    head=dict(
        type="LinearClsHead",
        num_classes=1000,
        in_channels=1280,
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
    ),
)



"""Configurable parameters for Draem anomaly task."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from attr import attrs

from otx.algorithms.anomaly.configs.base.configuration import BaseAnomalyConfig
from otx.algorithms.anomaly.configs.base.configuration_enums import EarlyStoppingMetrics
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    configurable_float,
    configurable_integer,
    selectable,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle


@attrs
class DraemAnomalyBaseConfig(BaseAnomalyConfig):
    """Configurable parameters for DRAEM anomaly classification task."""

    header = string_attribute("Configuration for Draem")
    description = header

    @attrs
    class LearningParameters(BaseAnomalyConfig.LearningParameters):
        """Parameters that can be tuned using HPO."""

        header = string_attribute("Learning Parameters")
        description = header

        train_batch_size = configurable_integer(
            default_value=8,
            min_value=1,
            max_value=512,
            header="Batch size",
            description="The number of training samples seen in each iteration of training. Increasing this value "
            "improves training time and may make the training more stable. A larger batch size has higher "
            "memory requirements.",
            warning="Increasing this value may cause the system to use more memory than available, "
            "potentially causing out of memory errors, please update with caution.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        lr = configurable_float(
            default_value=0.0001,
            header="Learning Rate",
            min_value=1e-4,
            max_value=1,
            description="Learning rate used for optimizing the network.",
        )

        @attrs
        class EarlyStoppingParameters(ParameterGroup):
            """Early stopping parameters."""

            header = string_attribute("Early Stopping Parameters")
            description = header

            metric = selectable(
                default_value=EarlyStoppingMetrics.IMAGE_ROC_AUC,
                header="Early Stopping Metric",
                description="The metric used to determine if the model should stop training",
            )

            patience = configurable_integer(
                default_value=20,
                min_value=1,
                max_value=100,
                header="Early Stopping Patience",
                description="Number of epochs to wait for an improvement in the monitored metric. If the metric has "
                "not improved for this many epochs, the training will stop and the best model will be "
                "returned.",
                warning="Setting this value too low might lead to underfitting. Setting the value too high will "
                "increase the training time and might lead to overfitting.",
                affects_outcome_of=ModelLifecycle.TRAINING,
            )

        early_stopping = add_parameter_group(EarlyStoppingParameters)

        max_epochs = configurable_integer(
            default_value=700,
            header="Max Epochs",
            min_value=1,
            max_value=700,
            description="Maximum number of epochs to train the model for.",
            warning="Training for very few epochs might lead to poor performance. If Early Stopping is enabled then "
            "increasing the value of max epochs might not lead to desired result.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    learning_parameters = add_parameter_group(LearningParameters)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmdeploy import is_mmdeploy_enabled
from .utils import numpy_2_list, sync_batchnorm_2_batchnorm

__all__ = [
    "is_mmdeploy_enabled",
    "sync_batchnorm_2_batchnorm",
    "numpy_2_list",
]



"""Tiling Pipeline of Resnet model for Instance-Seg Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

dataset_type = "CocoDataset"

img_size = (1344, 800)

tile_cfg = dict(
    tile_size=400, min_area_ratio=0.9, overlap_ratio=0.2, iou_threshold=0.45, max_per_img=1500, filter_empty_gt=True
)

img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

train_pipeline = [
    dict(type="Resize", img_scale=img_size, keep_ratio=False),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="Pad", size_divisor=32),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels", "gt_masks"]),
]

test_pipeline = [
    dict(
        type="MultiScaleFlipAug",
        img_scale=img_size,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **img_norm_cfg),
            dict(type="Pad", size_divisor=32),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    )
]

__dataset_type = "CocoDataset"
__data_root = "data/coco/"

__samples_per_gpu = 4

train_dataset = dict(
    type="ImageTilingDataset",
    dataset=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_train.json",
        img_prefix=__data_root + "images/train",
        pipeline=[
            dict(type="LoadImageFromFile"),
            dict(type="LoadAnnotations", with_bbox=True, with_mask=True),
        ],
    ),
    pipeline=train_pipeline,
    **tile_cfg
)

val_dataset = dict(
    type="ImageTilingDataset",
    dataset=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_val.json",
        img_prefix=__data_root + "images/val",
        test_mode=True,
        pipeline=[dict(type="LoadImageFromFile")],
    ),
    pipeline=test_pipeline,
    **tile_cfg
)

test_dataset = dict(
    type="ImageTilingDataset",
    dataset=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_test.json",
        img_prefix=__data_root + "images/test",
        test_mode=True,
        pipeline=[dict(type="LoadImageFromFile")],
    ),
    pipeline=test_pipeline,
    **tile_cfg
)


data = dict(
    samples_per_gpu=__samples_per_gpu, workers_per_gpu=2, train=train_dataset, val=val_dataset, test=test_dataset
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# FIXME: comment out for the time being
# due to annoying nncf warning messages
#
#  # flake8: noqa
#  try:
#      import mmcls as mmcls_import_test
#  except ImportError:
#      pass
#  else:
#      from . import mmcls
#
#  try:
#      import mmdet as mmdet_import_test
#  except ImportError:
#      pass
#  else:
#      from . import mmdet
#
#  try:
#      import mmseg as mmseg_import_test
#  except ImportError:
#      pass
#  else:
#      from . import mmseg



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.segmentation.tasks import SegmentationTrainTask
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class MockSegmentationTrainingTask(SegmentationTrainTask):
    def __init__(self):
        pass


class TestSegmentationTrainTaskInputParamsValidation:
    @staticmethod
    def model():
        model_configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="header", description="description"),
            label_schema=LabelSchemaEntity(),
        )
        return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)

    @e2e_pytest_unit
    def test_train_task_train_input_params_validation(self):
        """
        <b>Description:</b>
        Check SegmentationTrainTask object "train" method input parameters validation

        <b>Input data:</b>
        SegmentationTrainTask object, "train" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "train" method
        """
        task = MockSegmentationTrainingTask()
        correct_values_dict = {
            "dataset": DatasetEntity(),
            "output_model": self.model(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "output_model" parameter
            ("output_model", unexpected_str),
            # Unexpected string is specified as "train_parameters" parameter
            ("train_parameters", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.train,
        )

    @e2e_pytest_unit
    def test_train_task_save_model_input_params_validation(self):
        """
        <b>Description:</b>
        Check SegmentationTrainTask object "save_model" method input parameters validation

        <b>Input data:</b>
        SegmentationTrainTask object, "model" non-ModelEntity object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "save_model" method
        """
        task = MockSegmentationTrainingTask()
        with pytest.raises(ValueError):
            task.save_model("unexpected string")  # type: ignore



# Copyright (C) 2018-2021 OpenMMLab
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.cnn import build_activation_layer
from mmdet.models.builder import HEADS, build_loss
from mmdet.models.dense_heads.ssd_head import SSDHead
from mmdet.models.losses import smooth_l1_loss
from torch import nn


@HEADS.register_module()
class CustomSSDHead(SSDHead):
    def __init__(
        self,
        *args,
        bg_loss_weight=-1.0,
        loss_cls=dict(
            type="CrossEntropyLoss",
            use_sigmoid=False,
            reduction="none",
            loss_weight=1.0,
        ),
        loss_balancing=False,
        **kwargs
    ):

        super().__init__(*args, **kwargs)
        self.loss_cls = build_loss(loss_cls)
        self.bg_loss_weight = bg_loss_weight
        self.loss_balancing = loss_balancing
        if self.loss_balancing:
            self.loss_weights = torch.nn.Parameter(torch.FloatTensor(2))
            for i in range(2):
                self.loss_weights.data[i] = 0.0

    # TODO: remove this internal method
    # _init_layers of CustomSSDHead(this) and of SSDHead(parent)
    # Initialize almost the same model structure.
    # However, there are subtle differences
    # Theses differences make `load_state_dict_pre_hook()` go wrong
    def _init_layers(self):
        """Initialize layers of the head."""
        self.cls_convs = nn.ModuleList()
        self.reg_convs = nn.ModuleList()

        act_cfg = self.act_cfg.copy()
        act_cfg.setdefault("inplace", True)
        for in_channel, num_base_priors in zip(self.in_channels, self.num_base_priors):
            if self.use_depthwise:
                activation_layer = build_activation_layer(act_cfg)

                self.reg_convs.append(
                    nn.Sequential(
                        nn.Conv2d(in_channel, in_channel, kernel_size=3, padding=1, groups=in_channel),
                        nn.BatchNorm2d(in_channel),
                        activation_layer,
                        nn.Conv2d(in_channel, num_base_priors * 4, kernel_size=1, padding=0),
                    )
                )
                self.cls_convs.append(
                    nn.Sequential(
                        nn.Conv2d(in_channel, in_channel, kernel_size=3, padding=1, groups=in_channel),
                        nn.BatchNorm2d(in_channel),
                        activation_layer,
                        nn.Conv2d(in_channel, num_base_priors * self.cls_out_channels, kernel_size=1, padding=0),
                    )
                )
            else:
                self.reg_convs.append(nn.Conv2d(in_channel, num_base_priors * 4, kernel_size=3, padding=1))
                self.cls_convs.append(
                    nn.Conv2d(in_channel, num_base_priors * self.cls_out_channels, kernel_size=3, padding=1)
                )

    def loss_single(
        self,
        cls_score,
        bbox_pred,
        anchor,
        labels,
        label_weights,
        bbox_targets,
        bbox_weights,
        num_total_samples,
    ):
        """Compute loss of a single image.

        Args:
            cls_score (Tensor): Box scores for eachimage
                Has shape (num_total_anchors, num_classes).
            bbox_pred (Tensor): Box energies / deltas for each image
                level with shape (num_total_anchors, 4).
            anchors (Tensor): Box reference for each scale level with shape
                (num_total_anchors, 4).
            labels (Tensor): Labels of each anchors with shape
                (num_total_anchors,).
            label_weights (Tensor): Label weights of each anchor with shape
                (num_total_anchors,)
            bbox_targets (Tensor): BBox regression targets of each anchor
                weight shape (num_total_anchors, 4).
            bbox_weights (Tensor): BBox regression loss weights of each anchor
                with shape (num_total_anchors, 4).
            num_total_samples (int): If sampling, num total samples equal to
                the number of total anchors; Otherwise, it is the number of
                positive anchors.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """

        # Re-weigting BG loss
        label_weights = label_weights.reshape(-1)
        if self.bg_loss_weight >= 0.0:
            neg_indices = labels == self.num_classes
            label_weights = label_weights.clone()
            label_weights[neg_indices] = self.bg_loss_weight

        loss_cls_all = self.loss_cls(cls_score, labels, label_weights)
        if len(loss_cls_all.shape) > 1:
            loss_cls_all = loss_cls_all.sum(-1)
        # FG cat_id: [0, num_classes -1], BG cat_id: num_classes
        pos_inds = ((labels >= 0) & (labels < self.num_classes)).nonzero(as_tuple=False).reshape(-1)
        neg_inds = (labels == self.num_classes).nonzero(as_tuple=False).view(-1)

        num_pos_samples = pos_inds.size(0)
        num_neg_samples = self.train_cfg.neg_pos_ratio * num_pos_samples
        if num_neg_samples > neg_inds.size(0):
            num_neg_samples = neg_inds.size(0)
        topk_loss_cls_neg, _ = loss_cls_all[neg_inds].topk(num_neg_samples)
        loss_cls_pos = loss_cls_all[pos_inds].sum()
        loss_cls_neg = topk_loss_cls_neg.sum()
        loss_cls = (loss_cls_pos + loss_cls_neg) / num_total_samples

        if self.reg_decoded_bbox:
            # When the regression loss (e.g. `IouLoss`, `GIouLoss`)
            # is applied directly on the decoded bounding boxes, it
            # decodes the already encoded coordinates to absolute format.
            bbox_pred = self.bbox_coder.decode(anchor, bbox_pred)

        loss_bbox = smooth_l1_loss(
            bbox_pred,
            bbox_targets,
            bbox_weights,
            beta=self.train_cfg.smoothl1_beta,
            avg_factor=num_total_samples,
        )
        return loss_cls[None], loss_bbox

    def loss(self, *args, **kwargs):
        losses = super().loss(*args, **kwargs)
        losses_cls = losses["loss_cls"]
        losses_bbox = losses["loss_bbox"]

        if self.loss_balancing:
            losses_cls, losses_bbox = self._balance_losses(losses_cls, losses_bbox)

        return dict(loss_cls=losses_cls, loss_bbox=losses_bbox)

    def _balance_losses(self, losses_cls, losses_reg):
        loss_cls = sum(_loss.mean() for _loss in losses_cls)
        loss_cls = torch.exp(-self.loss_weights[0]) * loss_cls + 0.5 * self.loss_weights[0]

        loss_reg = sum(_loss.mean() for _loss in losses_reg)
        loss_reg = torch.exp(-self.loss_weights[1]) * loss_reg + 0.5 * self.loss_weights[1]

        return (loss_cls, loss_reg)



"""This module implements the Polygon Shape entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# Conflict with Isort
# pylint: disable=wrong-import-order

import datetime
import warnings
from operator import attrgetter
from typing import List, Optional

from shapely.geometry import Polygon as shapely_polygon

from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import Shape, ShapeType
from otx.api.utils.time_utils import now


class Point:
    """This class defines a Point with an X and Y coordinate.

    Multiple points can be used to represent a Polygon
    """

    __slots__ = ["x", "y"]

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __repr__(self):
        """String representation of the point."""
        return f"Point({self.x}, {self.y})"

    def __eq__(self, other):
        """Checks if two points have the same x and y coordinates."""
        if isinstance(other, Point):
            return self.x == other.x and self.y == other.y
        return False

    def normalize_wrt_roi(self, roi_shape: Rectangle) -> "Point":
        """The inverse of denormalize_wrt_roi_shape.

        Transforming Polygon from the `roi` coordinate system to the normalized coordinate system.
        This is used when the tasks want to save the analysis results.

        For example in Detection -> Segmentation pipeline, the analysis results of segmentation
        needs to be normalized to the roi (bounding boxes) coming from the detection.

        Args:
            roi_shape (Point): the shape of the roi
        """
        roi_shape = roi_shape.clip_to_visible_region()
        width = roi_shape.width
        height = roi_shape.height
        x1 = roi_shape.x1
        y1 = roi_shape.y1
        return Point(x=self.x * width + x1, y=self.y * height + y1)

    def denormalize_wrt_roi_shape(self, roi_shape: Rectangle) -> "Point":
        """The inverse of normalize_wrt_roi_shape.

        Transforming Polygon from the normalized coordinate system to the `roi` coordinate system.
        This is used to pull ground truth during training process of the tasks.
        Examples given in the Shape implementations.

        Args:
            roi_shape (Rectangle): the shape of the roi
        """
        roi_shape = roi_shape.clip_to_visible_region()

        return Point(
            x=(self.x - roi_shape.x1) / roi_shape.width,
            y=(self.y - roi_shape.y1) / roi_shape.height,
        )


class Polygon(Shape):
    """Represents a polygon formed by a list of coordinates.

    NB Freehand drawings are also stored as polygons.

    Args:
        points: list of Point's forming the polygon
        modification_date: last modified date
    """

    # pylint: disable=too-many-arguments; Requires refactor
    def __init__(
        self,
        points: List[Point],
        modification_date: Optional[datetime.datetime] = None,
    ):
        modification_date = now() if modification_date is None else modification_date
        super().__init__(
            shape_type=ShapeType.POLYGON,
            modification_date=modification_date,
        )

        if len(points) == 0:
            raise ValueError("Cannot create polygon with no points")

        self.points = points

        self.min_x = min(points, key=attrgetter("x")).x
        self.max_x = max(points, key=attrgetter("x")).x
        self.min_y = min(points, key=attrgetter("y")).y
        self.max_y = max(points, key=attrgetter("y")).y

        is_valid = True
        for (x, y) in [(self.min_x, self.min_y), (self.max_x, self.max_y)]:
            is_valid = is_valid and self._validate_coordinates(x, y)
        if not is_valid:
            points_str = "; ".join(str(p) for p in self.points)
            warnings.warn(
                f"{type(self).__name__} coordinates are invalid : {points_str}",
                UserWarning,
            )

    def __repr__(self):
        """String representation of the polygon."""
        return (
            f"Polygon(len(points)={len(self.points)},"
            f" min_x={self.min_x}, max_x={self.max_x}, min_y={self.min_y}, max_y={self.max_y})"
        )

    def __eq__(self, other):
        """Compares if the polygon has the same points and modification date."""
        if isinstance(other, Polygon):
            return self.points == other.points and self.modification_date == other.modification_date
        return False

    def __hash__(self):
        """Returns hash of the Polygon object."""
        return hash(str(self))

    def normalize_wrt_roi_shape(self, roi_shape: Rectangle) -> "Polygon":
        """Transforms from the `roi` coordinate system to the normalized coordinate system.

        This function is the inverse of ``denormalize_wrt_roi_shape``.

        Example:
            Assume we have Polygon `p1` which lives in the top-right quarter of a 2D space.
            The 2D space where `p1` lives in is an `roi` living in the top-left quarter of the normalized coordinate
            space. This function returns Polygon `p1` expressed in the normalized coordinate space.

            >>> from otx.api.entities.annotation import Annotation
            >>> from otx.api.entities.shapes.rectangle import Rectangle
            >>> p1 = Polygon(points=[Point(x=0.5, y=0.0), Point(x=0.75, y=0.2), Point(x=0.6, y=0.1)])
            >>> roi = Rectangle(x1=0.0, x2=0.5, y1=0.0, y2=0.5)
            >>> normalized = p1.normalize_wrt_roi_shape(roi_shape)
            >>> normalized
            Polygon(, len(points)=3)

        Args:
            roi_shape: Region of Interest

        Returns:
            New polygon in the image coordinate system
        """
        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        points = [p.normalize_wrt_roi(roi_shape) for p in self.points]
        return Polygon(points=points)

    def denormalize_wrt_roi_shape(self, roi_shape: Rectangle) -> "Polygon":
        """Transforming shape from the normalized coordinate system to the `roi` coordinate system.

        This function is the inverse of ``normalize_wrt_roi_shape``

        Example:
            Assume we have Polygon `p1` which lives in the top-right quarter of the normalized coordinate space.
            The `roi` is a rectangle living in the half right of the normalized coordinate space.
            This function returns Polygon `p1` expressed in the coordinate space of `roi`. (should return top-half)

            Polygon denormalized to a rectangle as ROI

            >>> from otx.api.entities.shapes.rectangle import Rectangle
            >>> from otx.api.entities.annotation import Annotation
            >>> p1 = Polygon(points=[Point(x=0.5, y=0.0), Point(x=0.75, y=0.2), Point(x=0.6, y=0.1)])
            >>> roi = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=1.0)  # the half-right
            >>> normalized = p1.denormalize_wrt_roi_shape(roi_shape)
            >>> normalized
            Polygon(, len(points)=3)

        Args:
            roi_shape: Region of Interest

        Returns:
            New polygon in the ROI coordinate system
        """
        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        points = [p.denormalize_wrt_roi_shape(roi_shape) for p in self.points]
        return Polygon(points=points)

    def _as_shapely_polygon(self) -> shapely_polygon:
        """Returns the Polygon object as a shapely polygon which is used for calculating intersection between shapes."""
        return shapely_polygon([(point.x, point.y) for point in self.points])

    def get_area(self) -> float:
        """Returns the approximate area of the shape.

        Area is a value between 0 and 1, computed by converting the Polygon to a shapely polygon and reading the
        `.area` property.

        NOTE: This method should not be relied on for exact area computation. The area is approximate, because shapes
        are continuous, but pixels are discrete.

        Example:

            >>> Polygon(points=[Point(x=0.0, y=0.5), Point(x=0.5, y=0.5), Point(x=0.75, y=0.75)]).get_area()
            0.0625

        Returns:
            area of the shape
        """
        return self._as_shapely_polygon().area



"""Unit Test for otx.algorithms.action.adapters.mmaction.utils.config_utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import defaultdict
from typing import Any, Dict

import numpy as np

from otx.algorithms.action.adapters.mmaction.data import OTXActionDetDataset
from otx.algorithms.action.adapters.mmaction.utils import det_eval
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit

FULL_BOX = np.array([[0, 0, 1, 1]])


class MockDataInfoProxy(OTXActionDetDataset._DataInfoProxy):
    """Mock clsass for data proxy in OTXActionDetDataset."""

    def __init__(self):
        self.video_infos = [
            {"img_key": "video_0,0", "gt_bboxes": FULL_BOX, "gt_labels": np.array([[0, 1, 0, 0]])},
            {"img_key": "video_0,1", "gt_bboxes": FULL_BOX, "gt_labels": np.array([[0, 0, 1, 0]])},
            {"img_key": "video_0,2", "gt_bboxes": FULL_BOX, "gt_labels": np.array([[0, 0, 0, 1]])},
        ]

    def __getitem__(self, index: int) -> Dict[str, Any]:
        return self.video_infos[index]


@e2e_pytest_unit
def test_det_eval() -> None:
    """Test for det_eval function.

    <Step>
        1. Generate sample predictions: Tuple[defaultdict]
        2. Generate sample labels: List[LabelEntity]
        3. Generate mock video_infos
    """

    pred_bboxes = defaultdict()
    pred_bboxes["video_0,0"] = [FULL_BOX[0], FULL_BOX[0], FULL_BOX[0]]
    pred_bboxes["video_0,1"] = [FULL_BOX[0], FULL_BOX[0], FULL_BOX[0]]
    pred_bboxes["video_0,2"] = [FULL_BOX[0], FULL_BOX[0], FULL_BOX[0]]
    pred_labels = defaultdict()
    pred_labels["video_0,0"] = [1, 2, 3]
    pred_labels["video_0,1"] = [1, 2, 3]
    pred_labels["video_0,2"] = [1, 2, 3]
    pred_confs = defaultdict()
    pred_confs["video_0,0"] = [1, 0, 0]
    pred_confs["video_0,1"] = [0, 1, 0]
    pred_confs["video_0,2"] = [0, 0, 1]
    predictions = (pred_bboxes, pred_labels, pred_confs)

    labels = [
        LabelEntity(name="0", domain=Domain.ACTION_DETECTION, id=ID(1)),
        LabelEntity(name="1", domain=Domain.ACTION_DETECTION, id=ID(2)),
        LabelEntity(name="2", domain=Domain.ACTION_DETECTION, id=ID(3)),
    ]

    video_infos = MockDataInfoProxy()
    custom_classes = [0, 1, 2, 3]

    out = det_eval(predictions, "mAP", labels, video_infos, None, True, custom_classes)
    assert out["mAP@0.5IOU"] == 1.0



"""Configurable parameters for anomaly classification task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from sys import maxsize

from attr import attrs

from otx.algorithms.anomaly.configs.base.configuration_enums import (
    POTQuantizationPreset,
)
from otx.api.configuration import ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    boolean_attribute,
    configurable_boolean,
    configurable_integer,
    selectable,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle


@attrs
class BaseAnomalyConfig(ConfigurableParameters):
    """Base OTX configurable parameters for anomaly classification task."""

    header = string_attribute("Configuration for an anomaly classification task")
    description = header

    @attrs
    class LearningParameters(ParameterGroup):
        """Parameters that can be tuned using HPO."""

        header = string_attribute("Learning Parameters")
        description = header

        train_batch_size = configurable_integer(
            default_value=32,
            min_value=1,
            max_value=512,
            header="Batch size",
            description="The number of training samples seen in each iteration of training. Increasing this value "
            "improves training time and may make the training more stable. A larger batch size has higher "
            "memory requirements.",
            warning="Increasing this value may cause the system to use more memory than available, "
            "potentially causing out of memory errors, please update with caution.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    @attrs
    class DatasetParameters(ParameterGroup):
        """Parameters related to dataloader."""

        header = string_attribute("Dataset Parameters")
        description = header

        num_workers = configurable_integer(
            default_value=8,
            min_value=0,
            max_value=36,
            header="Number of workers",
            description="Increasing this value might improve training speed however it might cause out of memory "
            "errors. If the number of workers is set to zero, data loading will happen in the main "
            "training thread.",
        )

    @attrs
    class POTParameters(ParameterGroup):
        """Training parameters for post-training optimization."""

        header = string_attribute("POT Parameters")
        description = header
        visible_in_ui = boolean_attribute(False)

        preset = selectable(
            default_value=POTQuantizationPreset.PERFORMANCE,
            header="Preset",
            description="Quantization preset that defines quantization scheme",
        )

        stat_subset_size = configurable_integer(
            header="Number of data samples",
            description="Number of data samples used for post-training optimization",
            default_value=300,
            min_value=1,
            max_value=maxsize,
        )

    @attrs
    class NNCFOptimization(ParameterGroup):
        """Parameters for NNCF optimization."""

        header = string_attribute("Optimization by NNCF")
        description = header

        enable_quantization = configurable_boolean(
            default_value=True,
            header="Enable quantization algorithm",
            description="Enable quantization algorithm",
        )

        enable_pruning = configurable_boolean(
            default_value=False,
            header="Enable filter pruning algorithm",
            description="Enable filter pruning algorithm",
        )

        pruning_supported = configurable_boolean(
            default_value=False,
            header="Whether filter pruning is supported",
            description="Whether filter pruning is supported",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    learning_parameters = add_parameter_group(LearningParameters)
    dataset = add_parameter_group(DatasetParameters)
    pot_parameters = add_parameter_group(POTParameters)
    nncf_optimization = add_parameter_group(NNCFOptimization)



"""This module implements the ID entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Optional, Union

from bson import ObjectId


class ID(str):
    """An identifier for objects that can be persisted in repositories.

    Usually the creation of IDs is handled by the repositories. In that case objects are initialized with an empty ID()
        >>> str(ID())
        ''

    Args:
        Optional[Union[str, ObjectId]]: value of the identifier
    """

    # Instead of using composition, we directly subclass the str builtin type.
    # This allows us to reuse most of the builtin functions of the str type,
    # and avoid the overhead of Python function calls for operations such as hashing and comparison.
    # In Python 3.6, using this approach, hash and equality are about 5 times faster than using composition.
    # Since the str type is immutable, we cannot handle the object construction arguments in __init__
    # as we would normally and instead have to use __new__.
    # The __init__ function is still left for typing correctness and so that Sphinx can get the prototype of the class.

    def __new__(cls, representation: Optional[Union[str, ObjectId]] = None):
        """Creates a new ID object."""
        if representation is None:
            representation = ""
        elif isinstance(representation, ObjectId):
            representation = str(representation)
        else:
            representation = str(representation).strip().lower()

        # Ignore typing error because Mypy does not support calling super().__new__ with a parameter.
        return super().__new__(cls, representation)  # type: ignore

    # See comment block above
    # pylint: disable=W0231
    def __init__(self, representation: Optional[Union[str, ObjectId]] = None):
        pass

    # This property name is there to allow automatic mapping between ID <> IDMessage
    # It should be the same as the argument in init
    @property
    def representation(self):
        """Returns the value of the identifier."""
        return self

    def __repr__(self):
        """Returns the representation of the identifier."""
        return f"ID({self})"



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import numpy as np
import pytest

from otx.algorithms.common.adapters.mmcv.hooks import OTXLoggerHook
from otx.algorithms.common.tasks import BaseTask
from otx.algorithms.detection.tasks import DetectionTrainTask
from otx.api.configuration.helper import create
from otx.api.entities.metrics import NullPerformance, Performance, ScoreMetric
from otx.api.entities.model_template import TaskType, parse_model_template
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_DET_TEMPLATE_DIR,
    generate_det_dataset,
    init_environment,
)


class TestOTXDetTaskTrain:
    @pytest.fixture(autouse=True)
    def setup(self, otx_model, tmp_dir_path) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_DET_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        task_env = init_environment(hyper_parameters, model_template)
        self.model = otx_model
        self.train_task = DetectionTrainTask(task_env, output_path=str(tmp_dir_path))

    @e2e_pytest_unit
    def test_save_model(self, mocker):
        """Test save_model method in OTXDetTaskTrain."""
        mocker.patch("torch.load", return_value="")
        self.train_task.save_model(self.model)

        assert self.model.get_data("weights.pth")
        assert self.model.get_data("label_schema.json")

    @e2e_pytest_unit
    def test_train(self, mocker):
        """Test train method in OTXDetTaskTrain."""
        self.dataset, _ = generate_det_dataset(task_type=TaskType.DETECTION)
        mock_lcurve_val = OTXLoggerHook.Curve()
        mock_lcurve_val.x = [0, 1]
        mock_lcurve_val.y = [0.1, 0.2]
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value={"final_ckpt": ""})
        self.train_task._learning_curves = {"val/mAP": mock_lcurve_val}
        mocker.patch.object(DetectionTrainTask, "save_model")

        fake_prediction = [[np.array([[0, 0, 32, 24, 0.55]], dtype=np.float32)]]
        fake_feature_vectors = [np.zeros((1, 1, 1))]
        fake_saliency_maps = [None]
        mocker.patch.object(
            DetectionTrainTask,
            "_infer_detector",
            return_value=(zip(fake_prediction, fake_feature_vectors, fake_saliency_maps), 1.0),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.f_measure.FMeasure", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics=["mAP"]
        )
        mocker.patch.object(MetricsHelper, "compute_f_measure", return_value=fake_metrics)

        self.train_task.train(self.dataset, self.model)

        mock_run_task.assert_called_once()
        assert self.model.performance != NullPerformance()
        assert self.model.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_cancel_training(self):
        """Test cancel_training method in OTXDetTaskTrain."""
        self.train_task.cancel_training()
        assert self.train_task._should_stop is True



"""NNCF utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .utils import (
    check_nncf_is_enabled,
    get_nncf_version,
    is_accuracy_aware_training_set,
    is_in_nncf_tracing,
    is_nncf_enabled,
    load_checkpoint,
    nncf_trace,
    no_nncf_trace,
)

__all__ = [
    "is_nncf_enabled",
    "check_nncf_is_enabled",
    "get_nncf_version",
    "load_checkpoint",
    "no_nncf_trace",
    "nncf_trace",
    "is_in_nncf_tracing",
    "is_accuracy_aware_training_set",
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp
from typing import Union

from otx.api.entities.metrics import Performance, ScoreMetric

# This string constant will be used as a special constant for a config field
# value to point that the field should be filled in tests' code by some default
# value specific for this field.
DEFAULT_FIELD_VALUE_FOR_USING_IN_TEST = "DEFAULT_FIELD_VALUE_FOR_USING_IN_TEST"


# This string constant will be used as a special constant for a config field value to point
# that the field should NOT be changed in tests -- its value should be taken
# from the template file or the config file of the model.
KEEP_CONFIG_FIELD_VALUE = "CONFIG"


# This is a constant for pointing usecase for reallife training tests
REALLIFE_USECASE_CONSTANT = "reallife"


# Constant for storing in dict-s with paths the root path
# that will be used for resolving relative paths.
ROOT_PATH_KEY = "_root_path"


def make_path_be_abs(some_val, root_path):
    assert isinstance(some_val, str), f"Wrong type of value: {some_val}, type={type(some_val)}"

    assert isinstance(root_path, str), f"Wrong type of root_path: {root_path}, type={type(root_path)}"

    # Note that os.path.join(a, b) == b if b is an absolute path
    return osp.join(root_path, some_val)


def make_paths_be_abs(some_dict, root_path):
    assert isinstance(some_dict, dict), f"Wrong type of value: {some_dict}, type={type(some_dict)}"

    assert isinstance(root_path, str), f"Wrong type of root_path: {root_path}, type={type(root_path)}"

    assert all(isinstance(v, str) for v in some_dict.values()), f"Wrong input dict {some_dict}"

    for k in list(some_dict.keys()):
        # Note that os.path.join(a, b) == b if b is an absolute path
        some_dict[k] = osp.join(root_path, some_dict[k])
    return some_dict


def performance_to_score_name_value(perf: Union[Performance, None]):
    """
    The method is intended to get main score info from Performance class
    """
    if perf is None:
        return None, None
    assert isinstance(perf, Performance)
    score = perf.score
    assert isinstance(score, ScoreMetric)
    assert isinstance(score.name, str) and score.name, f'Wrong score name "{score.name}"'
    return score.name, score.value



"""Classification models dynamic deploy config."""

_base_ = ["./base_classification_static.py"]

ir_config = dict(
    dynamic_axes={
        "data": {
            0: "batch",
            1: "channel",
            2: "height",
            3: "width",
        },
        "logits": {
            0: "batch",
        },
    }
)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch

from otx.mpa.modules.models.heads.custom_hierarchical_linear_cls_head import (
    CustomHierarchicalLinearClsHead,
)
from otx.mpa.modules.models.heads.custom_hierarchical_non_linear_cls_head import (
    CustomHierarchicalNonLinearClsHead,
)
from otx.mpa.modules.models.losses.asymmetric_loss_with_ignore import (
    AsymmetricLossWithIgnore,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCustomHierarchicalLinearClsHead:
    @pytest.fixture(autouse=True)
    def head_type(self) -> None:
        return CustomHierarchicalLinearClsHead

    @pytest.fixture(autouse=True)
    def setup(self, head_type) -> None:
        self.num_classes = 3
        self.head_dim = 5
        self.cls_heads_info = {
            "num_multiclass_heads": 1,
            "num_multilabel_classes": 1,
            "head_idx_to_logits_range": {0: (0, 2)},
            "num_single_label_classes": 2,
        }
        self.loss = dict(type="CrossEntropyLoss", use_sigmoid=False, reduction="mean", loss_weight=1.0)
        self.multilabel_loss = dict(type=AsymmetricLossWithIgnore.__name__, reduction="sum")
        self.default_head = head_type(
            self.num_classes,
            self.head_dim,
            hierarchical_info=self.cls_heads_info,
            loss=self.loss,
            multilabel_loss=self.multilabel_loss,
        )
        self.default_head.init_weights()
        self.default_input = torch.ones((2, self.head_dim))
        self.default_gt = torch.zeros((2, 2))

    @e2e_pytest_unit
    def test_forward(self) -> None:
        result = self.default_head.forward_train(self.default_input, self.default_gt)
        assert "loss" in result
        assert result["loss"] >= 0

    @e2e_pytest_unit
    def test_simple_test(self) -> None:
        result = self.default_head.simple_test(self.default_input)
        assert result[0].shape[0] == self.num_classes

    @e2e_pytest_unit
    def test_zero_classes(self, head_type) -> None:
        self.cls_heads_info["num_multiclass_heads"] = 0
        self.cls_heads_info["num_multilabel_classes"] = 0
        with pytest.raises(ValueError):
            head_type(
                self.num_classes,
                self.head_dim,
                hierarchical_info=self.cls_heads_info,
                loss=self.loss,
                multilabel_loss=self.multilabel_loss,
            )

    @e2e_pytest_unit
    def test_neg_classes(self, head_type) -> None:
        with pytest.raises(ValueError):
            head_type(
                -1,
                self.head_dim,
                hierarchical_info=self.cls_heads_info,
                loss=self.loss,
                multilabel_loss=self.multilabel_loss,
            )


class TestCustomHierarchicalNonLinearClsHead(TestCustomHierarchicalLinearClsHead):
    @pytest.fixture(autouse=True)
    def head_type(self) -> None:
        return CustomHierarchicalNonLinearClsHead

    @e2e_pytest_unit
    def test_dropout(self, head_type) -> None:
        head = head_type(self.num_classes, self.head_dim, dropout=True, hierarchical_info=self.cls_heads_info)
        head.init_weights()
        assert len(head.classifier) > len(self.default_head.classifier)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_rpn_head import MMOVRPNHead
from .mmov_ssd_head import MMOVSSDHead

# flake8: noqa
from .mmov_yolov3_head import MMOVYOLOV3Head



"""This module define the label entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import datetime
from enum import Enum, auto
from typing import Optional

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.utils.time_utils import now


class Domain(Enum):
    """Describes an algorithm domain like classification, detection, etc."""

    NULL = auto()
    CLASSIFICATION = auto()
    DETECTION = auto()
    SEGMENTATION = auto()
    ANOMALY_CLASSIFICATION = auto()
    ANOMALY_DETECTION = auto()
    ANOMALY_SEGMENTATION = auto()
    INSTANCE_SEGMENTATION = auto()
    ROTATED_DETECTION = auto()
    ACTION_CLASSIFICATION = auto()
    ACTION_DETECTION = auto()

    def __str__(self):
        """Returns Domain name."""
        return str(self.name)


class LabelEntity:
    """This represents a label. The Label is the object that the user annotates and the tasks predict.

    For example, a label with name "car" can be constructed as follows.

    >>> car = LabelEntity(name="car", domain=Domain.DETECTION)

    .. rubric:: About Empty Label

    In addition to representing the presence of a certain object, the label can also
    be used to represent the absence of objects in the image (or other media types).
    Such a label is referred to as empty label.
    The empty label is constructed as follows:

    >>> empty = LabelEntity(name="empty", domain=Domain.DETECTION, is_empty=True)

    Empty label is used to declare that there is nothing of interest inside this image.
    For example, let's assume a car detection project. During annotation process,
    for positive images (images with cars), the users are asked to annotate the images
    with bounding boxes with car label. However, when the user sees a negative image
    (no car), the user needs to annotate this image with an empty label.

    The empty label is particularly useful to distinguish images with no objects
    of interest from images that have not been annotated, especially in task-chain
    scenario. Let's assume car detection task that is followed with with another
    detection task which detects the driver inside the car. There are two issues here:

    1. The user can (intentionally or unintentionally) miss to annotate
        the driver inside a car.
    2. There is no driver inside the car.

    Without empty label, these two cases cannot be distinguished.
    This is why an empty label is introduced. The empty label makes an explicit
    distinction between missing annotations and "negative" images.

    Args:
        name: the name of the label
        domain: the algorithm domain this label is associated to
        color: the color of the label (See :class:`Color`)
        hotkey: key or combination of keys to select this label in the
            UI
        creation_date: the date time of the label creation
        is_empty: set to True if the label is an empty label.
        id: the ID of the label. Set to ID() so that a new unique ID
            will be assigned upon saving. If the argument is None, it
            will be set to ID()
        is_anomalous: boolean that indicates whether the label is the
            Anomalous label. Always set to False for non- anomaly
            projects.
    """

    # pylint: disable=redefined-builtin, too-many-instance-attributes, too-many-arguments; Requires refactor
    def __init__(
        self,
        name: str,
        domain: Domain,
        color: Optional[Color] = None,
        hotkey: str = "",
        creation_date: Optional[datetime.datetime] = None,
        is_empty: bool = False,
        id: Optional[ID] = None,
        is_anomalous: bool = False,
    ):
        id = ID() if id is None else id
        color = Color.random() if color is None else color
        creation_date = now() if creation_date is None else creation_date

        self._name = name
        self._color = color
        self._hotkey = hotkey
        self._domain = domain
        self._is_empty = is_empty
        self._creation_date = creation_date
        self.__id_ = id
        self.is_anomalous = is_anomalous

    @property
    def name(self):
        """Returns the label name."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def color(self) -> Color:
        """Returns the Color object for the label."""
        return self._color

    @color.setter
    def color(self, value):
        self._color = value

    @property
    def hotkey(self) -> str:
        """Returns the hotkey for the label."""
        return self._hotkey

    @hotkey.setter
    def hotkey(self, value):
        self._hotkey = value

    @property
    def domain(self):
        """Returns the algorithm domain associated to this label."""
        return self._domain

    @domain.setter
    def domain(self, value):
        self._domain = value

    @property
    def is_empty(self) -> bool:
        """Returns a boolean indicating if the label is an empty label."""
        return self._is_empty

    @property
    def creation_date(self) -> datetime.datetime:
        """Returns the creation date of the label."""
        return self._creation_date

    @property
    def id_(self) -> ID:
        """Returns the label id."""
        return self.__id_

    @id_.setter
    def id_(self, value: ID):
        self.__id_ = value

    @property
    def id(self) -> ID:
        """DEPRECATED."""
        return self.__id_

    @id.setter
    def id(self, value: ID):
        """DEPRECATED."""
        self.__id_ = value

    def __repr__(self):
        """String representation of the label."""
        return (
            f"LabelEntity({self.id_}, name={self.name}, hotkey={self.hotkey}, "
            f"domain={self.domain}, color={self.color}, is_anomalous={self.is_anomalous})"
        )

    def __eq__(self, other):
        """Returns True if the two labels are equal."""
        if isinstance(other, LabelEntity):
            return (
                self.id_ == other.id_
                and self.name == other.name
                and self.color == other.color
                and self.hotkey == other.hotkey
                and self.domain == other.domain
                and self.is_anomalous == other.is_anomalous
            )
        return False

    def __lt__(self, other):
        """Returns True if self.id < other.id."""
        if isinstance(other, LabelEntity):
            return self.id_ < other.id_
        return False

    def __gt__(self, other):
        """Returns True if self.id is greater than other.id."""
        if isinstance(other, LabelEntity):
            return self.id_ > other.id_
        return False

    def __hash__(self):
        """Returns hash of the label."""
        return hash(str(self))



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.parallel import is_module_wrapper
from mmcv.runner import HOOKS, Hook


@HOOKS.register_module()
class IBLossHook(Hook):
    def __init__(self, dst_classes):
        """Hook for IB loss.
        It passes the number of data per class and current epoch to IB loss class.

        Args:
            dst_classes (list): A list of classes including new_classes to be newly learned
        """
        self.cls_num_list = None
        self.dst_classes = dst_classes

    def before_train_epoch(self, runner):
        # get loss from model
        model_loss = self._get_model_loss(runner)

        # pass the number of data per class and current epoch to IB loss
        if runner.epoch == 0:
            dataset = runner.data_loader.dataset
            num_data = self._get_num_data(dataset)
            model_loss.update_weight(num_data)
        model_loss.cur_epoch = runner.epoch

    def _get_num_data(self, dataset):
        return [len(dataset.img_indices[data_cls]) for data_cls in self.dst_classes]

    def _get_model_loss(self, runner):
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        return model.head.compute_loss



"""Detection utils."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List

import numpy as np


def detection2array(detections: List) -> np.ndarray:
    """Convert list of OpenVINO Detection to a numpy array.

    Args:
        detections (List): List of OpenVINO Detection containing score, id, xmin, ymin, xmax, ymax

    Returns:
        np.ndarray: numpy array with [label, confidence, x1, y1, x2, y2]
    """
    scores = np.empty((0, 1), dtype=np.float32)
    labels = np.empty((0, 1), dtype=np.uint32)
    boxes = np.empty((0, 4), dtype=np.float32)
    for det in detections:
        if (det.xmax - det.xmin) * (det.ymax - det.ymin) < 1.0:
            continue
        scores = np.append(scores, [[det.score]], axis=0)
        labels = np.append(labels, [[det.id]], axis=0)
        boxes = np.append(
            boxes,
            [[float(det.xmin), float(det.ymin), float(det.xmax), float(det.ymax)]],
            axis=0,
        )
    detections = np.concatenate((labels, scores, boxes), -1)
    return detections



"""This module implements activation map."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from typing import Union

import cv2
import numpy as np


def get_actmap(
    saliency_map: np.ndarray,
    output_res: Union[tuple, list],
) -> np.ndarray:
    """Get activation map (heatmap)  from saliency map.

    It will return activation map from saliency map

    Args:
        saliency_map (np.ndarray): Saliency map with pixel values from 0-255
        output_res (Union[tuple, list]): Output resolution

    Returns:
        saliency_map (np.ndarray): [H, W, 3] colormap, more red means more salient

    """
    if len(saliency_map.shape) == 3:
        saliency_map = saliency_map[0]

    saliency_map = cv2.resize(saliency_map, output_res)
    saliency_map = cv2.applyColorMap(saliency_map, cv2.COLORMAP_JET)
    return saliency_map



"""Hyperband implementation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import math
import os
from os import path as osp
from typing import Any, Dict, List, Literal, Optional, Union

from scipy.stats.qmc import LatinHypercube

from otx.algorithms.common.utils.utils import (
    check_mode_input,
    check_not_negative,
    check_positive,
    left_vlaue_is_better,
)
from otx.hpo.hpo_base import HpoBase, Trial, TrialStatus

logger = logging.getLogger(__name__)


def _check_reduction_factor_value(reduction_factor: int):
    if reduction_factor < 2:
        raise ValueError("reduction_factor should be greater than 2.\n" f"your value : {reduction_factor}")


class AshaTrial(Trial):
    """ASHA trial class.

    Args:
        trial_id (Any): Id of the trial.
        configuration (Dict): Configuration for the trial.
        train_environment (Optional[Dict]): Train environment for the trial.
                                            For, example, subset ratio can be included. Defaults to None.
    """

    def __init__(self, trial_id: Any, configuration: Dict, train_environment: Optional[Dict] = None):
        super().__init__(trial_id, configuration, train_environment)
        self._rung: Optional[int] = None
        self._bracket: Optional[int] = None

    @property
    def rung(self):
        """Rung where the trial is included."""
        return self._rung

    @rung.setter
    def rung(self, val: int):
        """Setter for rung."""
        check_not_negative(val, "rung")
        self._rung = val

    @property
    def bracket(self):
        """Bracket where the trial is inlcuded."""
        return self._bracket

    @bracket.setter
    def bracket(self, val: int):
        """Setter for bracket."""
        check_not_negative(val, "bracket")
        self._bracket = val

    def save_results(self, save_path: str):
        """Save a result of the trial at 'save_path'."""
        results = {
            "id": self.id,
            "rung": self.rung,
            "configuration": self.configuration,
            "train_environment": self.train_environment,
            "score": self.score,
        }

        with open(save_path, "w", encoding="utf-8") as f:
            json.dump(results, f)


class Rung:
    """Rung class.

    Rung is in charge of selecting a trial to train and deciding which trial to promote to next rung in the bracket.

    Args:
        resource (Union[int, float]): Resource to use for training a trial.
                                      For example, something like epoch or iteration.
        num_required_trial (int): Necessary trials for the rung.
        reduction_factor (int): Decicdes how many trials to promote.
                                Only top 1 / reduction_factor of all trials can be promoted.
        rung_idx (int): Current rung index.
    """

    def __init__(
        self,
        resource: Union[int, float],
        num_required_trial: int,
        reduction_factor: int,
        rung_idx: int,
    ):
        check_positive(resource, "resource")
        check_positive(num_required_trial, "num_required_trial")
        _check_reduction_factor_value(reduction_factor)
        check_not_negative(rung_idx, "rung_idx")

        self._reduction_factor = reduction_factor
        self._num_required_trial = num_required_trial
        self._resource = resource
        self._trials: List[AshaTrial] = []
        self._rung_idx = rung_idx

    @property
    def num_required_trial(self):
        """Number of required trials for the rung."""
        return self._num_required_trial

    @property
    def resource(self):
        """Resource to use for training a trial."""
        return self._resource

    @property
    def rung_idx(self):
        """Current rung index."""
        return self._rung_idx

    def add_new_trial(self, trial: AshaTrial):
        """Add a new trial to the rung.

        Args:
            trial (AshaTrial): Trial to add

        Raises:
            RuntimeError: If no more trial is needed, raise an error.
        """
        if not self.need_more_trials():
            raise RuntimeError(f"{self.rung_idx} rung has already sufficient trials.")
        trial.iteration = self.resource
        trial.rung = self.rung_idx
        trial.status = TrialStatus.READY
        self._trials.append(trial)

    def get_best_trial(self, mode: str = "max") -> Optional[AshaTrial]:
        """Get best trial in the rung.

        Args:
            mode (str, optional): Decide which trial is better between having highest score or lowest score.
                                  Defaults to "max".

        Returns:
            Optional[AshaTrial]: Best trial. If there is no trial, return None.
        """
        check_mode_input(mode)
        best_score = None
        best_trial = None
        for trial in self._trials:
            if trial.rung != self.rung_idx:
                continue
            trial_score = trial.get_best_score(mode, self.resource)
            if trial_score is not None and (best_score is None or left_vlaue_is_better(trial_score, best_score, mode)):
                best_trial = trial
                best_score = trial_score

        return best_trial

    def need_more_trials(self) -> bool:
        """Check whether the rung needs more trials."""
        return self.num_required_trial > self.get_num_trials()

    def get_num_trials(self) -> int:
        """Number of trials the rung has."""
        return len(self._trials)

    def is_done(self) -> bool:
        """Check that the rung is done."""
        if self.need_more_trials():
            return False
        for trial in self._trials:
            if not trial.is_done():
                return False
        return True

    def get_trial_to_promote(self, asynchronous_sha: bool = False, mode: str = "max") -> Optional[AshaTrial]:
        """Get a trial to promote.

        Args:
            asynchronous_sha (bool, optional): Whether to operate SHA asynchronously. Defaults to False.
            mode (str, optional): Decide which trial is better between having highest score or lowest score.
                                  Defaults to "max".

        Returns:
            Optional[AshaTrial]: Trial to prmote. If there is no trial to promote, return None.
        """
        num_finished_trial = 0
        num_promoted_trial = 0
        best_score = None
        best_trial = None

        for trial in self._trials:
            if trial.rung == self._rung_idx:
                if trial.is_done() and trial.status != TrialStatus.RUNNING:
                    num_finished_trial += 1
                    trial_score = trial.get_best_score(mode, self.resource)
                    if best_score is None or left_vlaue_is_better(trial_score, best_score, mode):
                        best_trial = trial
                        best_score = trial_score
            else:
                num_promoted_trial += 1

        if asynchronous_sha:
            if (num_promoted_trial + num_finished_trial) // self._reduction_factor > num_promoted_trial:
                return best_trial
        else:
            if self.is_done() and self._num_required_trial // self._reduction_factor > num_promoted_trial:
                return best_trial

        return None

    def get_next_trial(self) -> Optional[AshaTrial]:
        """Get next trial to trian.

        Returns:
            Optional[AshaTrial]: Next trial to train. If there is no left trial to train, then return None.
        """
        for trial in self._trials:
            if not trial.is_done() and trial.status != TrialStatus.RUNNING:
                return trial
        return None


class Bracket:
    """Bracket class. It operates a single SHA using multiple rungs.

    Args:
        bracket_id (int): Bracket id.
        minimum_resource (Union[float, int]): Maximum resource to use for training a trial.
        maximum_resource (Union[float, int]): Minimum resource to use for training a trial.
        hyper_parameter_configurations (List[AshaTrial]): Hyper parameter configuration to try.
        reduction_factor (int): Decicdes how many trials to promote to next rung.
                                Only top 1 / reduction_factor of rung trials can be promoted.
        mode (str, optional): Decide which trial is better between having highest score or lowest score.
                              Defaults to "max".
        asynchronous_sha (bool, optional): Whether to operate SHA asynchronously. Defaults to True.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        bracket_id: int,
        minimum_resource: Union[float, int],
        maximum_resource: Union[float, int],
        hyper_parameter_configurations: List[AshaTrial],
        reduction_factor: int = 3,
        mode: str = "max",
        asynchronous_sha: bool = True,
    ):
        # pylint: disable=too-many-arguments
        check_positive(minimum_resource, "minimum_resource")
        _check_reduction_factor_value(reduction_factor)
        check_mode_input(mode)

        self._id = bracket_id
        self._minimum_resource = minimum_resource
        self.maximum_resource = maximum_resource
        self._reduction_factor = reduction_factor
        self._mode = mode
        self._asynchronous_sha = asynchronous_sha
        self._trials: Dict[int, AshaTrial] = {}
        self._rungs: List[Rung] = self._initialize_rungs(hyper_parameter_configurations)

    @property
    def id(self):
        """Bracket id."""
        return self._id

    @property
    def maximum_resource(self):
        """Maximum resource to use for training a trial."""
        return self._maximum_resource

    @maximum_resource.setter
    def maximum_resource(self, val: Union[float, int]):
        """Setter for maximum_resource."""
        check_positive(val, "maximum_resource")
        if val < self._minimum_resource:
            raise ValueError(
                "maxnum_resource should be greater than minimum_resource.\n"
                f"value to set : {val}, minimum_resource : {self._minimum_resource}"
            )
        if val == self._minimum_resource:
            logger.warning("maximum_resource is same with the minimum_resource.")

        self._maximum_resource = val

    @property
    def max_rung(self):
        """Number of rungs the bracket has."""
        return self.calcuate_max_rung_idx(self._minimum_resource, self.maximum_resource, self._reduction_factor)

    @staticmethod
    def calcuate_max_rung_idx(
        minimum_resource: Union[float, int], maximum_resource: Union[float, int], reduction_factor: int
    ) -> int:
        """Calculate the number of rungs the bracket needs.

        Args:
            minimum_resource (Union[float, int]): Minimum resource to use for training a trial.
            maximum_resource (Union[float, int]): Maximum resource to use for training a trial.
            reduction_factor (int): Decicdes how many trials to promote to next rung.
                                    Only top 1 / reduction_factor of rung trials can be promoted.

        Raises:
            ValueError: If minimum resource is lower than maximum resource, raise an error.

        Returns:
            int: The number of rungs the bracket needs.
        """
        check_positive(minimum_resource, "minimum_resource")
        check_positive(maximum_resource, "maximum_resource")
        check_positive(reduction_factor, "reduction_factor")
        if minimum_resource > maximum_resource:
            raise ValueError(
                "maximum_resource should be bigger than minimum_resource. "
                f"but minimum_resource : {minimum_resource} / maximum_resource : {maximum_resource}"
            )

        return math.ceil(math.log(maximum_resource / minimum_resource, reduction_factor))

    def _initialize_rungs(self, hyper_parameter_configurations: List[AshaTrial]):
        num_trials = len(hyper_parameter_configurations)
        minimum_num_trials = self._reduction_factor**self.max_rung
        if minimum_num_trials > num_trials:
            raise ValueError(
                "number of hyper_parameter_configurations is not enough. "
                f"minimum number is {minimum_num_trials}, but current number is {num_trials}. "
                "if you want to let them be, you can decrease needed number "
                "by increasing reduction factor or minimum resource."
            )

        rungs = [
            Rung(
                self._minimum_resource * (self._reduction_factor**idx),
                math.floor(num_trials * (self._reduction_factor**-idx)),
                self._reduction_factor,
                idx,
            )
            for idx in range(self.max_rung + 1)
        ]

        for new_trial in hyper_parameter_configurations[: rungs[0].num_required_trial]:
            new_trial.bracket = self.id
            rungs[0].add_new_trial(new_trial)
            self._trials[new_trial.id] = new_trial

        return rungs

    def _promote_trial_if_available(self, rung_idx: int):
        check_not_negative(rung_idx, "rung_idx")

        if self.max_rung <= rung_idx:
            return None

        best_trial = self._rungs[rung_idx].get_trial_to_promote(self._asynchronous_sha, self._mode)
        if best_trial is not None:
            self._rungs[rung_idx + 1].add_new_trial(best_trial)

        return best_trial

    def get_next_trial(self) -> Optional[AshaTrial]:
        """Get next trial to train.

        Returns:
            Optional[AshaTrial]: Next trial to train. There is no trial to train, then return None.
        """
        current_rung = self.max_rung
        while current_rung >= 0:
            next_sample = self._promote_trial_if_available(current_rung)
            if next_sample is not None:
                if next_sample.is_done():
                    if current_rung < self.max_rung - 1:
                        current_rung += 1
                    continue
                break

            next_sample = self._rungs[current_rung].get_next_trial()
            if next_sample is not None:
                break

            current_rung -= 1

        return next_sample

    def is_done(self) -> bool:
        """Check that the bracket is done.

        Returns:
            bool: Whether bracket is done or not.
        """
        return self._rungs[-1].is_done()

    def get_best_trial(self) -> Optional[AshaTrial]:
        """Get best trial in the bracket.

        Returns:
            Optional[AshaTrial]: Best trial in the bracket. If there is no trial to select, then return None.
        """
        if not self.is_done():
            logger.warning("Bracket is not done yet.")

        trial = None
        for rung in reversed(self._rungs):
            trial = rung.get_best_trial(self._mode)
            if trial is None:
                continue
            break

        return trial

    def save_results(self, save_path: str):
        """Save a bracket result to 'save_path'.

        Args:
            save_path (str): Path where to save a bracket result.
        """
        result = self._get_result()
        with open(osp.join(save_path, "rung_status.json"), "w", encoding="utf-8") as f:
            json.dump(result, f)

        for trial_id, trial in self._trials.items():
            trial.save_results(osp.join(save_path, f"{trial_id}.json"))

    def print_result(self):
        """Print a bracket result."""
        print("*" * 20, f"{self.id} bracket", "*" * 20)
        result = self._get_result()
        del result["rung_status"]
        for key, val in result.items():
            print(f"{key} : {val}")

        best_trial = self.get_best_trial()
        if best_trial is None:
            print("This bracket isn't started yet!\n")
            return
        print(
            f"best trial:\n"
            f"id : {best_trial.id} / score : {best_trial.get_best_score()} / config : {best_trial.configuration}"
        )

        print("all trials:")
        for trial in self._trials.values():
            print(f"id : {trial.id} / score : {trial.get_best_score()} / config : {trial.configuration}")
        print()

    def _get_result(self):
        return {
            "minimum_resource": self._minimum_resource,
            "maximum_resource": self.maximum_resource,
            "reduction_factor": self._reduction_factor,
            "mode": self._mode,
            "asynchronous_sha": self._asynchronous_sha,
            "num_trials": len(self._trials),
            "rung_status": [
                {
                    "rung_idx": rung.rung_idx,
                    "num_trial": rung.get_num_trials(),
                    "num_required_trial": rung.num_required_trial,
                    "resource": rung.resource,
                }
                for rung in self._rungs
            ],
        }


class HyperBand(HpoBase):
    """It implements the Asyncronous HyperBand scheduler with iterations only.

    Please refer the below papers for the detailed algorithm.

    [1] "Hyperband: A Novel Bandit-Based Approach to Hyperparameter Optimization", JMLR 2018
        https://arxiv.org/abs/1603.06560
        https://homes.cs.washington.edu/~jamieson/hyperband.html

    [2] "A System for Massively Parallel Hyperparameter Tuning", MLSys 2020
        https://arxiv.org/abs/1810.05934

    Args:
        minimum_resource (Union[float, int]): Minimum resource to use for training a trial. Defaults to None.
        reduction_factor (int, optional): Decicdes how many trials to promote to next rung.
                                          Only top 1 / reduction_factor of rung trials can be promoted. Defaults to 3.
        asynchronous_sha (bool, optional): Whether to operate SHA asynchronously. Defaults to True.
        asynchronous_bracket (bool, optional): Whether SHAs(brackets) are running parallelly or not.
                                               Defaults to True. Defaults to False.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        minimum_resource: Optional[Union[int, float]] = None,
        reduction_factor: int = 3,
        asynchronous_sha: bool = True,
        asynchronous_bracket: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if minimum_resource is not None:
            check_positive(minimum_resource, "minimum_resource")
        _check_reduction_factor_value(reduction_factor)

        self._next_trial_id = 0
        self._reduction_factor = reduction_factor
        self._minimum_resource = minimum_resource
        self._asynchronous_sha = asynchronous_sha
        self._asynchronous_bracket = asynchronous_bracket
        self._trials: Dict[str, AshaTrial] = {}
        self._brackets: Dict[int, Bracket] = {}

        if not self._need_to_find_resource_value():
            self._brackets = self._make_brackets()

    def _need_to_find_resource_value(self) -> bool:
        return self.maximum_resource is None or self._minimum_resource is None

    def _make_brackets(self) -> Dict[int, Bracket]:
        if self.expected_time_ratio is None:
            brackets_config = self._make_default_brackets_setting()
        else:
            brackets_config = self.auto_config()
        return self._make_brackets_as_config(brackets_config)

    def _calculate_bracket_resource(self, num_max_rung_trials: int, bracket_index: int) -> Union[int, float]:
        """Calculate how much resource is needed for the bracket given that resume is available."""
        num_trial = self._calculate_num_bracket_trials(num_max_rung_trials, bracket_index)
        minimum_resource = self.maximum_resource * (self._reduction_factor**-bracket_index)

        total_resource = 0
        num_rungs = (
            Bracket.calcuate_max_rung_idx(
                minimum_resource, self.maximum_resource, self._reduction_factor  # type: ignore
            )
            + 1
        )
        previous_resource = 0
        resource = minimum_resource
        for _ in range(num_rungs):
            total_resource += num_trial * (resource - previous_resource)
            num_trial //= self._reduction_factor
            previous_resource = resource
            resource *= self._reduction_factor

        return total_resource

    def _calculate_num_bracket_trials(self, num_max_rung_trials: int, bracket_index: int) -> int:
        return num_max_rung_trials * (self._reduction_factor**bracket_index)

    def _calculate_origin_num_trial_for_bracket(self, bracket_idx: int) -> int:
        return self._calculate_num_bracket_trials(self._get_num_max_rung_trials(bracket_idx), bracket_idx)

    def _get_num_max_rung_trials(self, bracket_idx: int) -> int:
        return math.floor((self._calculate_s_max() + 1) / (bracket_idx + 1))

    def _calculate_s_max(self) -> int:
        return math.floor(
            math.log(self.maximum_resource / self._minimum_resource, self._reduction_factor)  # type: ignore
        )

    def _make_default_brackets_setting(self) -> List[Dict[str, Any]]:
        # Bracket order is the opposite of order of paper's.
        # This is for running default hyper parmeters with abundant resource.
        brackets_setting = []
        for idx in range(self._calculate_s_max() + 1):
            brackets_setting.append(
                {"bracket_index": idx, "num_trials": self._calculate_origin_num_trial_for_bracket(idx)}
            )

        return brackets_setting

    def _make_brackets_as_config(self, brackets_settings: List[Dict]) -> Dict[int, Bracket]:
        brackets = {}
        total_num_trials = 0
        for bracket_setting in brackets_settings:
            total_num_trials += bracket_setting["num_trials"]
        reserved_trials = list(self._trials.values()) if self._trials else []
        if len(reserved_trials) > total_num_trials:
            reserved_trials = reserved_trials[:total_num_trials]
        configurations = self._make_new_hyper_parameter_configs(total_num_trials - len(reserved_trials))

        for bracket_setting in brackets_settings:
            bracket_idx = bracket_setting["bracket_index"]
            num_trial_to_initialize = bracket_setting["num_trials"]
            minimum_resource = self.maximum_resource * (self._reduction_factor**-bracket_idx)

            bracket_configurations = []
            for reserved_trial in reserved_trials:
                if (
                    reserved_trial.bracket is None and reserved_trial.get_progress() <= minimum_resource
                ) or reserved_trial.bracket == bracket_idx:
                    num_trial_to_initialize -= 1
                    bracket_configurations.append(reserved_trial)
                    if len(bracket_configurations) >= bracket_setting["num_trials"]:
                        break

            for selected_trial in bracket_configurations:
                reserved_trials.remove(selected_trial)

            bracket_configurations.extend(configurations[:num_trial_to_initialize])
            configurations = configurations[num_trial_to_initialize:]

            bracket = Bracket(
                bracket_idx,
                minimum_resource,
                self.maximum_resource,  # type: ignore
                bracket_configurations,
                self._reduction_factor,
                self.mode,
                self._asynchronous_sha,
            )
            brackets[bracket_idx] = bracket

        return brackets

    def _make_new_hyper_parameter_configs(self, num: int) -> List[AshaTrial]:
        check_not_negative(num, "num")

        hp_configs: List[AshaTrial] = []
        if num == 0:
            return hp_configs

        hp_configs.extend(self._get_prior_hyper_parameters(num))
        if num - len(hp_configs) > 0:
            hp_configs.extend(self._get_random_hyper_parameter(num - len(hp_configs)))

        return hp_configs

    def _get_prior_hyper_parameters(self, num_samples: int) -> List[AshaTrial]:
        hp_configs = []
        num_samples = min([num_samples, len(self.prior_hyper_parameters)])
        for _ in range(num_samples):
            hyper_parameter = self.prior_hyper_parameters.pop(0)
            hp_configs.append(self._make_trial(hyper_parameter))

        return hp_configs

    def _get_random_hyper_parameter(self, num_samples: int) -> List[AshaTrial]:
        hp_configs = []
        latin_hypercube = LatinHypercube(len(self.search_space))
        configurations = latin_hypercube.random(num_samples)
        for config in configurations:
            config_with_key = {key: config[idx] for idx, key in enumerate(self.search_space)}
            hp_configs.append(
                self._make_trial(self.search_space.convert_from_zero_one_scale_to_real_space(config_with_key))
            )

        return hp_configs

    def _make_trial(self, hyper_parameter: Dict) -> AshaTrial:
        trial_id = self._get_new_trial_id()
        trial = AshaTrial(trial_id, hyper_parameter, self._get_train_environment())
        self._trials[trial_id] = trial
        return trial

    def _get_new_trial_id(self) -> str:
        trial_id = self._next_trial_id
        self._next_trial_id += 1
        return str(trial_id)

    def _get_train_environment(self) -> Dict:
        train_environment = {"subset_ratio": self.subset_ratio}
        return train_environment

    def get_next_sample(self) -> Optional[AshaTrial]:
        """Get next trial to train.

        Returns:
            Optional[AshaTrial]: Next trial to train. If there is no trial to train, then return None.
        """
        if not self._brackets:
            return self._make_trial_to_estimate_resource()

        next_sample = None
        for bracket in self._brackets.values():
            if not bracket.is_done():
                next_sample = bracket.get_next_trial()
                if self._asynchronous_bracket and next_sample is None:
                    continue
                break

        return next_sample

    def _make_trial_to_estimate_resource(self) -> AshaTrial:
        """Trial to estimate a maximum resource or minimum resource."""
        trial = self._make_new_hyper_parameter_configs(1)[0]
        if self.maximum_resource is None:
            if len(self._trials) == 1:  # first trial to estimate
                trial.bracket = 0
                trial.iteration = self.num_full_iterations
            elif self._minimum_resource is not None:
                trial.iteration = self._minimum_resource
            else:
                trial.iteration = self.num_full_iterations
        else:
            trial.iteration = self.maximum_resource
        return trial

    def save_results(self):
        """Save a ASHA result."""
        for idx, bracket in self._brackets.items():
            save_path = osp.join(self.save_path, str(idx))
            os.makedirs(save_path, exist_ok=True)
            bracket.save_results(save_path)

    def auto_config(self) -> List[Dict[str, Any]]:
        """Configure ASHA automatically aligning with possible resource.

        Configure ASHA automatically. If resource is lesser than full ASHA, decrease ASHA scale.
        In contrast, resource is more than full ASHA, increase ASHA scale.

        Returns:
            List[Dict[str, Any]]: ASHA configuration. It's used to make brackets.
        """
        if self._trials:
            self._adjust_minimum_resource()
        if self._need_to_dcrease_hyerpband_scale():
            return self._decrease_hyperband_scale()
        return self._increase_hyperband_scale()

    def _adjust_minimum_resource(self):
        """Set meaningful minimum resource.

        Goal of this function is to avoid setting minimum resource too low
        to distinguish which trial is better.
        """
        if self.maximum_resource < self._reduction_factor:
            logger.debug("maximum_resource is lesser than reduction factor. adjusting minimum resource is skipped.")
            return

        trial = None
        for trial in self._trials.values():
            if trial.is_done():
                break
        if trial is None:
            logger.debug("There is no finished trial. adjusting minimum resource is skipped.")
            return

        cur_score = 0
        best_score = 0
        minimum_resource = 0
        for resource, score in trial.score.items():
            if resource > self.maximum_resource // self._reduction_factor:
                break
            cur_score = cur_score * 0.5 + score * 0.5
            if not left_vlaue_is_better(best_score, cur_score, self.mode):
                best_score = cur_score
                if minimum_resource == 0:
                    minimum_resource = resource
            else:
                minimum_resource = 0

        if minimum_resource == 0:
            minimum_resource = self.maximum_resource // self._reduction_factor
        self._minimum_resource = minimum_resource

    def _get_full_asha_resource(self) -> Union[int, float]:
        total_resource: Union[int, float] = 0
        for idx in range(self._calculate_s_max() + 1):
            num_max_rung_trials = self._get_num_max_rung_trials(idx)
            total_resource += self._calculate_bracket_resource(num_max_rung_trials, idx)

        return total_resource

    def _need_to_dcrease_hyerpband_scale(self) -> bool:
        """Check full ASHA resource exceeds expected_time_ratio."""
        if self.expected_time_ratio is None:
            return False

        return self._get_full_asha_resource() > self._get_expected_total_resource()

    def _decrease_hyperband_scale(self) -> List[Dict[str, Any]]:
        """Decrease Hyperband scale.

        From bracket which has biggest number of rung, check that it's resource exceeds expected_time_ratio
        if bracket is added. If not, bracket is added. If it does, check that number of trials for bracket
        can be reduced. if not, skip that bracket and check that next bracket can be added by same method.
        """
        brackets_setting: List = []
        total_resource: Union[int, float] = 0
        resource_upper_bound = self._get_expected_total_resource()

        reserved_resource: Union[int, float] = 0
        if self._trials:  # reserve resources for trials which should be run on bracket 0
            for trial in self._trials.values():
                if trial.bracket == 0:
                    reserved_resource += self.maximum_resource  # type: ignore
            total_resource += reserved_resource

        for idx in range(self._calculate_s_max(), -1, -1):
            if self._trials and idx == 0:
                total_resource -= reserved_resource

            origin_num_max_rung_trials = self._get_num_max_rung_trials(idx)
            for num_max_rung_trials in range(origin_num_max_rung_trials, 0, -1):
                bracket_resource = self._calculate_bracket_resource(num_max_rung_trials, idx)

                if total_resource + bracket_resource <= resource_upper_bound:
                    total_resource += bracket_resource
                    num_bracket_trials = self._calculate_num_bracket_trials(num_max_rung_trials, idx)
                    brackets_setting.insert(0, {"bracket_index": idx, "num_trials": num_bracket_trials})
                    break

        return brackets_setting

    def _get_expected_total_resource(self):
        if self.expected_time_ratio is None:
            raise ValueError("expected time ratio should be set to get expceted total resource")
        return (
            self.num_full_iterations
            * self.expected_time_ratio
            * self.acceptable_additional_time_ratio
            * self.num_workers
        )

    def _increase_hyperband_scale(self) -> List[Dict[str, Any]]:
        total_resource: Union[int, float] = 0
        bracket_status = {}
        s_max = self._calculate_s_max()

        # If all brackets can run more than one, then multiply number of trials on each bracket as many as possible
        sum_unit_resource: Union[int, float] = 0
        for idx in range(s_max + 1):
            num_max_rung_trials = self._get_num_max_rung_trials(idx)
            unit_resource = self._calculate_bracket_resource(1, idx)
            sum_unit_resource += unit_resource
            bracket_status[idx] = {"num_max_rung_trials": num_max_rung_trials, "unit_resource": unit_resource}
            total_resource += num_max_rung_trials * unit_resource

        maximum_reseource = self._get_expected_total_resource()
        available_num_trials = int((maximum_reseource - total_resource) // sum_unit_resource)

        for idx in range(s_max + 1):
            bracket_status[idx]["num_max_rung_trials"] += available_num_trials
        total_resource += sum_unit_resource * available_num_trials

        # add trials to brackets from big index as many as possible
        while True:
            update_flag = False
            for idx in range(s_max, -1, -1):
                if total_resource + bracket_status[idx]["unit_resource"] < maximum_reseource:
                    total_resource += bracket_status[idx]["unit_resource"]
                    bracket_status[idx]["num_max_rung_trials"] += 1
                    update_flag = True

            if not update_flag:
                break

        # set brackets setting
        brackets_setting = []
        for idx in range(s_max + 1):
            brackets_setting.append(
                {
                    "bracket_index": idx,
                    "num_trials": self._calculate_num_bracket_trials(
                        bracket_status[idx]["num_max_rung_trials"], idx  # type: ignore
                    ),
                }
            )

        return brackets_setting

    def _get_used_resource(self) -> Union[int, float]:
        used_resource: Union[int, float] = 0
        for trial in self._trials.values():
            used_resource += trial.get_progress()

        return used_resource

    def get_progress(self) -> Union[int, float]:
        """Get current progress of ASHA."""
        if self.is_done():
            return 1

        if self.expected_time_ratio is None:
            total_resource = self._get_full_asha_resource()
        else:
            total_resource = self._get_expected_total_resource()

        progress = self._get_used_resource() / total_resource

        return min(progress, 0.99)

    def report_score(
        self, score: Union[float, int], resource: Union[float, int], trial_id: str, done: bool = False
    ) -> Literal[TrialStatus.STOP, TrialStatus.RUNNING]:
        """Report a score to ASHA.

        Args:
            score (Union[float, int]): Score to report.
            resource (Union[float, int]): Resource used to get score.
            trial_id (str): Trial id.
            done (bool, optional): Whether training trial is done. Defaults to False.

        Returns:
            Literal[TrialStatus.STOP, TrialStatus.RUNNING]: Decide whether to continue training or not.
        """
        trial = self._trials[trial_id]
        if done:
            if self.maximum_resource is None:
                self.maximum_resource = trial.get_progress()
                self.num_full_iterations = self.maximum_resource
                if not self._need_to_find_resource_value():
                    self._brackets = self._make_brackets()
            trial.finalize()
        else:
            trial.register_score(score, resource)
            if self._minimum_resource is None:
                self._minimum_resource = min(trial.score.keys())
                if not self._need_to_find_resource_value():
                    self._brackets = self._make_brackets()
            if trial.is_done() or trial.bracket is None:
                return TrialStatus.STOP

        return TrialStatus.RUNNING

    def is_done(self) -> bool:
        """Check that the ASHA is done.

        Returns:
            bool: Whether ASHA is done.
        """
        if not self._brackets:
            return False
        for bracket in self._brackets.values():
            if not bracket.is_done():
                return False
        return True

    def get_best_config(self) -> Optional[Dict[str, Any]]:
        """Get best configuration in ASHA.

        Returns:
            Optional[Dict[str, Any]]: Best configuration in ASHA. If there is no trial to select, return None.
        """
        best_score = None
        best_trial = None

        for trial in self._trials.values():
            score = trial.get_best_score()
            if score is not None and (best_score is None or left_vlaue_is_better(score, best_score, self.mode)):
                best_score = score
                best_trial = trial

        if best_trial is None:
            return None
        return {"id": best_trial.id, "config": best_trial.configuration}

    def print_result(self):
        """Print a ASHA result."""
        print(
            "HPO(ASHA) result summary\n"
            f"Best config : {self.get_best_config()}.\n"
            f"Hyper band runs {len(self._brackets)} brackets.\n"
            "Brackets summary:"
        )
        for bracket in self._brackets.values():
            bracket.print_result()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.cls.stage import ClsStage
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class SemiSLClsStage(ClsStage):
    """Patch config to support semi supervised learning for object Cls"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_data(self, cfg, data_cfg, training, **kwargs):
        """Patch cfg.data."""
        super().configure_data(cfg, data_cfg, training, **kwargs)
        # Set unlabeled data hook
        if training:
            if cfg.data.get("unlabeled", False) and cfg.data.unlabeled.get("otx_dataset", False):
                self.configure_unlabeled_dataloader(cfg, self.distributed)



"""This module contains the definition for the `validate` function within the configuration helper.

This function can be used to validate the values of a OTX configuration object, checking that each of the parameter
values in the configuration are within their allowed bounds or options.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import attr

from otx.api.configuration.elements import ParameterGroup


def _validate_inner(config: ParameterGroup):
    """Recursive method that performs validation on all parameters within a parameter group.

    Uses recursion to validate parameters living inside nested groups.
    """
    attr.validate(config)
    if config.groups:
        for group_name in config.groups:
            group = getattr(config, group_name)
            _validate_inner(group)


def validate(config: ParameterGroup) -> bool:
    """Validate a configuration object.

    Args:
        config: Configuration to validate

    Returns:
        True if config is valid
    """
    _validate_inner(config)
    return True



"""Collection Pipeline for classification task."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import copy
import tempfile
from typing import Any, Dict, List

import numpy as np
from mmcls.datasets import PIPELINES
from mmcls.datasets.pipelines import Compose
from mmcv.utils.registry import build_from_cfg
from PIL import Image, ImageFilter
from torchvision import transforms as T

from otx.algorithms.common.utils.data import get_image
from otx.api.utils.argument_checks import check_input_parameters_type

_CACHE_DIR = tempfile.TemporaryDirectory(prefix="img-cache-")  # pylint: disable=consider-using-with

# TODO: refactoring to common modules
# TODO: refactoring to Sphinx style.


@PIPELINES.register_module()
class LoadImageFromOTXDataset:
    """Pipeline element that loads an image from a OTX Dataset on the fly.

    Can do conversion to float 32 if needed.
    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the image
        results['dataset_id']: id of the dataset to which the item belongs
        results['index']: index of the item in the dataset

    :param to_float32: optional bool, True to convert images to fp32. defaults to False
    """

    @check_input_parameters_type()
    def __init__(self, to_float32: bool = False):
        self.to_float32 = to_float32

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadImageFromOTXDataset."""
        # Get image (possibly from cache)
        img = get_image(results, _CACHE_DIR.name, to_float32=self.to_float32)
        shape = img.shape

        assert img.shape[0] == results["height"], f"{img.shape[0]} != {results['height']}"
        assert img.shape[1] == results["width"], f"{img.shape[1]} != {results['width']}"

        filename = f"Dataset item index {results['index']}"
        results["filename"] = filename
        results["ori_filename"] = filename
        results["img"] = img
        results["img_shape"] = shape
        results["ori_shape"] = shape
        # Set initial values for default meta_keys
        results["pad_shape"] = shape
        num_channels = 1 if len(shape) < 3 else shape[2]
        results["img_norm_cfg"] = dict(
            mean=np.zeros(num_channels, dtype=np.float32),
            std=np.ones(num_channels, dtype=np.float32),
            to_rgb=False,
        )
        results["img_fields"] = ["img"]

        return results


@PIPELINES.register_module()
class RandomAppliedTrans:
    """Randomly applied transformations.

    :param transforms: List of transformations in dictionaries
    :param p: Probability, defaults to 0.5
    """

    def __init__(self, transforms: List, p: float = 0.5):
        t = [build_from_cfg(t, PIPELINES) for t in transforms]  # pylint: disable=invalid-name
        self.trans = T.RandomApply(t, p=p)

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of RandomAppliedTrans.

        :param results: Inputs to be transformed.
        """
        return self.trans(results)

    def __repr__(self):
        """Set repr of RandomAppliedTrans."""
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module
class OTXColorJitter(T.ColorJitter):
    """Wrapper for ColorJitter in torchvision.transforms.

    Use this instead of mmcls's because there is no `hue` parameter in mmcls ColorJitter.
    """

    def __call__(self, results):
        """Callback function of OTXColorJitter.

        :param results: Inputs to be transformed.
        """
        results["img"] = np.array(self.forward(Image.fromarray(results["img"])))
        return results


@PIPELINES.register_module()
class GaussianBlur:
    """Gaussian blur augmentation in SimCLR https://arxiv.org/abs/2002.05709.

    :param sigma_min: Minimum value of sigma of gaussian filter.
    :param sigma_max: Maximum value of sigma of gaussian filter.
    """

    @check_input_parameters_type()
    def __init__(self, sigma_min: float, sigma_max: float):
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of GaussianBlur.

        :param results: Inputs to be transformed.
        """
        for key in results.get("img_fields", ["img"]):
            img = Image.fromarray(results[key])
            sigma = np.random.uniform(self.sigma_min, self.sigma_max)
            results[key] = np.array(img.filter(ImageFilter.GaussianBlur(radius=sigma)))
        return results

    def __repr__(self):
        """Set repr of GaussianBlur."""
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class PILImageToNDArray:
    """Pipeline element that converts an PIL image into numpy array.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['img']: PIL type image in data pipeline.

    Args:
        keys (list[str]): list to support multiple image converting from PIL to NDArray.
    """

    def __init__(self, keys=None):
        self.keys = keys

    def __call__(self, results):
        """Callback function of PILImageToNDArray."""
        for key in self.keys:
            img = results[key]
            img = np.asarray(img)
            results[key] = img
        return results

    def __repr__(self):
        """Repr function of PILImageToNDArray."""
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class PostAug:
    """Pipeline element that postaugments for mmcls.

    PostAug copies current augmented image and apply post augmentations for given keys.
    For example, if we apply PostAug(keys=dict(img_strong=strong_pipeline),
    PostAug will copy current augmented image and apply strong pipeline.
    Post augmented image will be saved at results["img_strong"].

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['img']: PIL type image in data pipeline.

    Args:
        keys (dict): keys to apply postaugmentaion. ex) dict(img_strong=strong_pipeline)
    """

    def __init__(self, keys: dict):
        self.pipelines = {key: Compose(pipeline) for key, pipeline in keys.items()}

    def __call__(self, results):
        """Callback function of PostAug."""
        for key, pipeline in self.pipelines.items():
            results[key] = pipeline(copy.deepcopy(results))["img"]
            results["img_fields"].append(key)
        return results

    def __repr__(self):
        """Repr function of PostAug."""
        repr_str = self.__class__.__name__
        return repr_str



import pytest

from otx.mpa.seg.semisl.stage import SemiSLSegStage
from otx.mpa.seg.stage import SegStage
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit

SEMISL_RECIPE_CONFIG_PATH = "otx/recipes/stages/segmentation/semisl.py"


class TestOTXSegStage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(SEMISL_RECIPE_CONFIG_PATH)
        self.stage = SemiSLSegStage(name="", mode="train", config=cfg, common_cfg=None, index=0)

    @e2e_pytest_unit
    def test_configure_data(self, mocker):
        mock_ul_dataloader = mocker.patch.object(SegStage, "configure_unlabeled_dataloader")
        fake_semisl_data_cfg = {"data": {"unlabeled": {"otx_dataset": "foo"}}}
        self.stage.configure_data(self.stage.cfg, True, fake_semisl_data_cfg)

        mock_ul_dataloader.assert_called_once()

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        fake_model_cfg = {"model": {"type": "", "task_adapt": True}}
        self.stage.cfg.merge_from_dict(fake_model_cfg)
        mock_remove_hook = mocker.patch("otx.mpa.seg.semisl.stage.remove_custom_hook")
        self.stage.configure_task(self.stage.cfg, True)

        assert "task_adapt" not in self.stage.cfg.model
        mock_remove_hook.assert_called_once()



__dataset_type = ""
__data_root = ""
__pipeline = ""

data = dict(
    train=dict(
        type="RepeatDataset", times=1, dataset=dict(type=__dataset_type, data_root=__data_root, pipeline=__pipeline)
    ),
    val=dict(type=__dataset_type, data_root=__data_root, pipeline=__pipeline),
    test=dict(type=__dataset_type, data_root=__data_root, pipeline=__pipeline),
    unlabeled=dict(type=__dataset_type, data_root=__data_root, pipeline=__pipeline),
)



"""Adapters for nncf support."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa

from otx.core.patcher import Patcher

from .compression import (
    AccuracyAwareLrUpdater,
    get_nncf_metadata,
    get_uncompressed_model,
    is_checkpoint_nncf,
    is_state_nncf,
)
from .patches import (
    nncf_trace_context,
    nncf_trace_wrapper,
    nncf_train_step,
    no_nncf_trace_wrapper,
)
from .utils import (
    check_nncf_is_enabled,
    get_nncf_version,
    is_accuracy_aware_training_set,
    is_in_nncf_tracing,
    is_nncf_enabled,
    no_nncf_trace,
)

NNCF_PATCHER = Patcher()


__all__ = [
    "NNCF_PATCHER",
    "AccuracyAwareLrUpdater",
    "check_nncf_is_enabled",
    "get_nncf_metadata",
    "get_nncf_version",
    "get_uncompressed_model",
    "is_accuracy_aware_training_set",
    "is_checkpoint_nncf",
    "is_in_nncf_tracing",
    "is_state_nncf",
    "no_nncf_trace",
    "nncf_trace_context",
    "nncf_train_step",
    "no_nncf_trace_wrapper",
    "nncf_trace_wrapper",
    "is_nncf_enabled",
]



"""Model training tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=too-many-locals

from pathlib import Path

from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.cli.manager import ConfigManager
from otx.cli.utils.hpo import run_hpo
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_binary, read_label_schema, save_model_data
from otx.cli.utils.multi_gpu import MultiGPUManager
from otx.cli.utils.parser import (
    add_hyper_parameters_sub_parser,
    get_parser_and_hprams_data,
)
from otx.core.data.adapter import get_dataset_adapter


def get_args():
    """Parses command line arguments."""
    parser, hyper_parameters, params = get_parser_and_hprams_data()

    parser.add_argument(
        "--train-data-roots",
        help="Comma-separated paths to training data folders.",
    )
    parser.add_argument(
        "--val-data-roots",
        help="Comma-separated paths to validation data folders.",
    )
    parser.add_argument(
        "--unlabeled-data-roots",
        help="Comma-separated paths to unlabeled data folders",
    )
    parser.add_argument(
        "--unlabeled-file-list",
        help="Comma-separated paths to unlabeled file list",
    )
    parser.add_argument(
        "--load-weights",
        help="Load model weights from previously saved checkpoint.",
    )
    parser.add_argument(
        "--resume-from",
        help="Resume training from previously saved checkpoint",
    )
    parser.add_argument(
        "--save-model-to",
        help="Location where trained model will be stored.",
    )
    parser.add_argument(
        "--work-dir",
        help="Location where the intermediate output of the training will be stored.",
        default=None,
    )
    parser.add_argument(
        "--enable-hpo",
        action="store_true",
        help="Execute hyper parameters optimization (HPO) before training.",
    )
    parser.add_argument(
        "--hpo-time-ratio",
        default=4,
        type=float,
        help="Expected ratio of total time to run HPO to time taken for full fine-tuning.",
    )
    parser.add_argument(
        "--gpus",
        type=str,
        help="Comma-separated indices of GPU. \
              If there are more than one available GPU, then model is trained with multi GPUs.",
    )
    parser.add_argument(
        "--rdzv-endpoint",
        type=str,
        default="localhost:0",
        help="Rendezvous endpoint for multi-node training.",
    )
    parser.add_argument(
        "--base-rank",
        type=int,
        default=0,
        help="Base rank of the current node workers.",
    )
    parser.add_argument(
        "--world-size",
        type=int,
        default=0,
        help="Total number of workers in a worker group.",
    )
    parser.add_argument(
        "--data",
        type=str,
        default=None,
        help="The data.yaml path want to use in train task.",
    )

    sub_parser = add_hyper_parameters_sub_parser(parser, hyper_parameters, return_sub_parser=True)
    # TODO: Temporary solution for cases where there is no template input
    override_param = [f"params.{param[2:].split('=')[0]}" for param in params if param.startswith("--")]
    if not hyper_parameters and "params" in params:
        if "params" in params:
            params = params[params.index("params") :]
            for param in params:
                if param == "--help":
                    print("Without template configuration, hparams information is unknown.")
                elif param.startswith("--"):
                    sub_parser.add_argument(
                        f"{param}",
                        dest=f"params.{param[2:]}",
                    )
    return parser.parse_args(), override_param


def main():  # pylint: disable=too-many-branches
    """Main function that is used for model training."""
    args, override_param = get_args()

    config_manager = ConfigManager(args, workspace_root=args.work_dir, mode="train")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # Creates a workspace if it doesn't exist.
    if not config_manager.check_workspace():
        config_manager.build_workspace(new_workspace_path=args.work_dir)

    # Auto-Configuration for Dataset configuration
    config_manager.configure_data_config(update_data_yaml=config_manager.check_workspace())
    dataset_config = config_manager.get_dataset_config(subsets=["train", "val", "unlabeled"])
    dataset_adapter = get_dataset_adapter(**dataset_config)
    dataset, label_schema = dataset_adapter.get_otx_dataset(), dataset_adapter.get_label_schema()

    # Get classes for Task, ConfigurableParameters and Dataset.
    template = config_manager.template
    task_class = get_impl_class(template.entrypoints.base)

    # Update Hyper Parameter Configs
    hyper_parameters = config_manager.get_hyparams_config(override_param=override_param)

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=hyper_parameters,
        label_schema=label_schema,
        model_template=template,
    )

    if args.load_weights or args.resume_from:
        ckpt_path = args.resume_from if args.resume_from else args.load_weights
        model_adapters = {
            "weights.pth": ModelAdapter(read_binary(ckpt_path)),
            "resume": bool(args.resume_from),
        }

        if (Path(ckpt_path).parent / "label_schema.json").exists():
            model_adapters.update(
                {"label_schema.json": ModelAdapter(label_schema_to_bytes(read_label_schema(ckpt_path)))}
            )

        environment.model = ModelEntity(
            train_dataset=dataset,
            configuration=environment.get_model_configuration(),
            model_adapters=model_adapters,
        )

    # FIXME: Need to align output results & Current HPO use save_model_to.parent
    if "save_model_to" not in args or not args.save_model_to:
        args.save_model_to = str(config_manager.workspace_root / "models")
    if args.enable_hpo:
        environment = run_hpo(args, environment, dataset, config_manager.data_config)

    task = task_class(task_environment=environment, output_path=args.work_dir)

    if args.gpus:
        multigpu_manager = MultiGPUManager(main, args.gpus, args.rdzv_endpoint, args.base_rank, args.world_size)
        if template.task_type in (TaskType.ACTION_CLASSIFICATION, TaskType.ACTION_DETECTION):
            print("Multi-GPU training for action tasks isn't supported yet. A single GPU will be used for a training.")
        elif (
            multigpu_manager.is_available()
            and not template.task_type.is_anomaly  # anomaly tasks don't use this way for multi-GPU training
        ):
            multigpu_manager.setup_multi_gpu_train(task.project_path, hyper_parameters if args.enable_hpo else None)

    output_model = ModelEntity(dataset, environment.get_model_configuration())

    task.train(dataset, output_model, train_parameters=TrainParameters())

    save_model_data(output_model, args.save_model_to)
    print(f"[*] Save Model to: {args.save_model_to}")

    if config_manager.data_config["val_subset"]["data_root"]:
        validation_dataset = dataset.get_subset(Subset.VALIDATION)
        predicted_validation_dataset = task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=False),
        )

        resultset = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        task.evaluate(resultset)
        assert resultset.performance is not None
        print(resultset.performance)

    task.cleanup()

    if args.gpus:
        multigpu_manager.finalize()

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



"""This module implements the TrainParameters entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Union

import networkx as nx

from otx.api.entities.interfaces.graph_interface import IGraph


class Graph(IGraph):
    """The concrete implementation of IGraph. This implementation is using networkx library.

    Args:
        directed (bool): set to True if the graph is a directed graph.
    """

    def __init__(self, directed: bool = False):
        self._graph: Union[nx.Graph, nx.MultiDiGraph] = nx.Graph() if not directed else nx.MultiDiGraph()
        self.directed = directed

    def get_graph(self) -> Union[nx.Graph, nx.MultiDiGraph]:
        """Get the underlying NetworkX graph."""
        return self._graph

    def set_graph(self, graph: Union[nx.Graph, nx.MultiDiGraph]):
        """Set the underlying NetworkX graph."""
        self._graph = graph

    def add_edge(self, node1, node2, edge_value=None):
        """Adds edge between node1 and node2."""
        # pylint: disable=arguments-differ
        self._graph.add_edge(node1, node2, value=edge_value)

    def num_nodes(self) -> int:
        """Returns the number of nodes in the graph."""
        return self._graph.number_of_nodes()

    def add_node(self, node):
        """Adds node to the graph."""
        if node not in self._graph.nodes:
            self._graph.add_node(node)

    def has_edge_between(self, node1, node2):
        """Returns True if there is an edge between node1 and node2."""
        return node1 in self.neighbors(node2)

    def neighbors(self, node):
        """Returns neighbors of `label`.

        Note: when `node` does not exist in the graph an empty list is returned
        """
        try:
            result = list(self._graph.neighbors(node))
        except nx.NetworkXError:
            result = []
        return result

    def find_out_edges(self, node):
        """Returns the edges that have `node` as a destination."""
        # pylint: disable=no-member
        if node not in self._graph.nodes:
            raise KeyError(f"The node `{node}` is not part of the graph")

        if isinstance(self._graph, nx.MultiDiGraph):
            return self._graph.out_edges(node)
        return []

    def find_in_edges(self, node):
        """Returns the edges that have `node` as a source."""
        # pylint: disable=no-member
        if node not in self._graph.nodes:
            raise KeyError(f"The node `{node}` is not part of the graph")

        if isinstance(self._graph, nx.MultiDiGraph):
            return self._graph.in_edges(node)
        return []

    def find_cliques(self):
        """Returns cliques in the graph."""
        return nx.algorithms.clique.find_cliques(self._graph)

    @property
    def nodes(self):
        """Returns the nodes in the graph."""
        return self._graph.nodes

    @property
    def edges(self):
        """Returns all the edges in the graph."""
        if isinstance(self._graph, nx.MultiDiGraph):
            all_edges = self._graph.edges(keys=True, data=True)
        else:
            all_edges = self._graph.edges(data=True)
        return all_edges

    @property
    def num_labels(self):
        """Returns the number of labels in the graph."""
        return nx.convert_matrix.to_numpy_matrix(self._graph).shape[0]

    def remove_edges(self, node1, node2):
        """Removes edges between both the nodes."""
        self._graph.remove_edge(node1, node2)

    def remove_node(self, node):
        """Remove node from graph.

        Args:
            node: node to remove
        """
        self._graph.remove_node(node)

    def descendants(self, parent):
        """Returns descendants.

        (children and children of children, etc.) of `parent`.
        """
        try:
            edges = list(nx.edge_dfs(self._graph, parent, orientation="reverse"))
        except nx.exception.NetworkXError:
            edges = []
        return [edge[0] for edge in edges]

    def __eq__(self, other: object) -> bool:
        """Returns True if the two graphs are equal."""
        if isinstance(other, Graph):
            return (
                self.directed == other.directed
                and self._graph.nodes == other._graph.nodes
                and self._graph.edges == other._graph.edges
            )
        return False


class MultiDiGraph(Graph):
    """Multi Dimensional implementation of a Graph."""

    def __init__(self) -> None:
        super().__init__(directed=True)

    def topological_sort(self):
        """Returns a generator of nodes in topologically sorted order."""
        return nx.topological_sort(self._graph)



"""Training reporting.

.. automodule:: otx.api.usecases.reporting.callback
   :members:
   :undoc-members:

.. automodule:: otx.api.usecases.reporting.time_monitor_callback
   :members:
   :undoc-members:

"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .callback import Callback
from .time_monitor_callback import TimeMonitorCallback

__all__ = ["Callback", "TimeMonitorCallback"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os
import os.path as osp
from tempfile import TemporaryDirectory

import cv2
import mmcv
import numpy as np
import pytest

from otx.algorithms.segmentation.adapters.mmseg.utils.data_utils import (
    add_labels,
    check_labels,
    create_annotation_from_hard_seg_map,
    create_pseudo_masks,
    load_dataset_items,
)
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def label_entity(name="test label", id=None) -> LabelEntity:
    return LabelEntity(name=name, id=id, domain=Domain.SEGMENTATION)


def generate_random_single_image(filename: str, width: int = 10, height: int = 10) -> None:
    img: np.ndarray = np.uint8(np.random.random((height, width, 3)) * 255)
    cv2.imwrite(filename, img)


class TestMMSegDataUtilsValidation:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:

        self.hard_seg_map: np.ndarray = np.array([[1, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])
        self.labels: list = [label_entity("class_0", id="00000000"), label_entity("class_1", id="00000001")]

    @e2e_pytest_unit
    def test_create_annotation_from_hard_seg_map(self) -> None:
        annotations: list = create_annotation_from_hard_seg_map(self.hard_seg_map, self.labels)

        assert len(annotations) == 1
        assert annotations[0].get_labels()[0].label.name == "class_0"

    @e2e_pytest_unit
    def test_check_labels(self) -> None:
        check_labels(cur_labels=self.labels, new_labels=[("class_1", None), ("class_0", None)])
        # function doesn't return anything, but throws exeption in case of failure
        assert True

    @e2e_pytest_unit
    def test_add_labels(self) -> None:
        add_labels(cur_labels=self.labels, new_labels=[("class_2", None)])

        assert len(self.labels) == 3

    @e2e_pytest_unit
    def test_create_pseudo_masks_fh_mode(self, mocker) -> None:
        mocker.patch("cv2.imread", return_value=np.array([[1, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]]))
        mocker.patch("cv2.imwrite")
        mocker.patch("os.listdir", return_value=["image.jpg"])

        ann_file_path: str = "ann_file_path"
        data_root_dir: str = "data_root_dir"

        # ann_file_path dir should not exist
        create_pseudo_masks(ann_file_path, data_root_dir)

        assert osp.exists(osp.join(ann_file_path, "meta.json"))

        os.remove(f"{ann_file_path}/meta.json")
        os.rmdir(ann_file_path)

    @e2e_pytest_unit
    def test_load_dataset_items(self) -> None:

        tmp_dir: TemporaryDirectory = TemporaryDirectory()

        generate_random_single_image(osp.join(tmp_dir.name, "image.jpg"))
        generate_random_single_image(osp.join(tmp_dir.name, "image.png"))
        fake_json_file: str = osp.join(tmp_dir.name, "meta.json")
        mmcv.dump({"labels_map": []}, fake_json_file)

        # ann_file_path dir should exist
        dataset_items: list = load_dataset_items(tmp_dir.name, tmp_dir.name)

        assert len(dataset_items) == 1



"""Fast RCNN for Action Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from mmcv import ConfigDict
from mmcv.runner import load_checkpoint
from mmdet.models import DETECTORS
from mmdet.models.builder import build_detector
from mmdet.models.detectors import FastRCNN
from torch import nn

from otx.algorithms.action.configs.detection.base.faster_rcnn_config import (
    faster_rcnn,
    faster_rcnn_pretrained,
)


class ONNXPool3D(nn.Module):
    """3D pooling method for onnx export.

    ONNX dose not support dynamic pooling, therefore pooling operation should be changed into static way
    """

    def __init__(self, dim, pool_type):
        super().__init__()
        self.dim = dim
        if pool_type == "avg":
            self.pool = nn.functional.avg_pool3d
        else:
            self.pool = nn.functional.max_pool3d

    def forward(self, x):
        """Forward method."""
        size_array = [int(s) for s in x.size()[2:]]
        if self.dim == "temporal":
            kernel_size = [size_array[0], 1, 1]
        else:
            kernel_size = [1, size_array[1], size_array[2]]
        return self.pool(x, kernel_size)


@DETECTORS.register_module()
class AVAFastRCNN(FastRCNN):
    """Implementation of `Fast R-CNN` for Action Detection.

    Add forward_infer function for inference without pre-proposals
    """

    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(self, backbone, roi_head, train_cfg, test_cfg, neck=None, pretrained=None):
        super(FastRCNN, self).__init__(
            backbone=backbone,
            neck=neck,
            roi_head=roi_head,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            pretrained=pretrained,
        )
        self.detector = None

    def add_detector(self):
        """Add Person Detector for inference.

        Action classification backbone + Fast RCNN structure use pre-proposals instead outputs from person detector
        This saves training and evaluation time. However, this method doesn't support inference for the dataset
        without pre-proposals. Therefore, when the model infers to dataset without pre-proposal, person detector
        should be added to action detector.
        """
        detector = ConfigDict(faster_rcnn)
        self.detector = build_detector(detector)
        ckpt = load_checkpoint(self.detector, faster_rcnn_pretrained, map_location="cpu")
        self.detector.CLASSES = ckpt["meta"]["CLASSES"]
        if self.detector.CLASSES[0] != "person":
            raise Exception(
                f"Person detector should have person as the first category, but got {self.detector.CLASSES}"
            )

    def patch_pools(self):
        """Patch pooling functions for ONNX export.

        AVAFastRCNN's bbox head has pooling funcitons, which contain dynamic shaping.
        This funciton changes those pooling functions from dynamic shaping to static shaping.
        """
        self.roi_head.bbox_head.temporal_pool = ONNXPool3D("temporal", self.roi_head.bbox_head.temporal_pool_type)
        self.roi_head.bbox_head.spatial_pool = ONNXPool3D("spatial", self.roi_head.bbox_head.spatial_pool_type)

    def forward_infer(self, imgs, img_metas):
        """Forward function for inference without pre-proposal."""
        clip_len = imgs[0].shape[2]
        img = imgs[0][:, :, int(clip_len / 2), :, :]
        det_bboxes, det_labels = self.detector.onnx_export(img, img_metas[0])
        prediction = [det_bboxes[0][det_labels[0] == 0]]
        prediction = self.forward_test(imgs, img_metas, proposals=[prediction])
        return prediction



"""BYOL (Bootstrap Your Own Latent) implementation for self-supervised learning.

Original papers:
- 'Bootstrap Your Own Latent: A New Approach to Self-Supervised Learning', https://arxiv.org/abs/2006.07733
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=missing-module-docstring, too-many-instance-attributes, unused-argument, unnecessary-pass, invalid-name  # noqa: E501
from collections import OrderedDict
from typing import Any, Dict, Optional

import torch
import torch.distributed as dist
from mmcls.models.builder import CLASSIFIERS, build_backbone, build_head, build_neck
from torch import nn

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@CLASSIFIERS.register_module()
class BYOL(nn.Module):
    """Implementation of 'Bootstrap Your Own Latent: A New Approach to \
    Self-Supervised Learning (https://arxiv.org/abs/2006.07733)'.

    Args:
        backbone (dict): Config dict for module of backbone ConvNet.
        neck (dict): Config dict for module of deep features to compact feature vectors.
            Default: None.
        head (dict): Config dict for module of loss functions. Default: None.
        pretrained (str, optional): Path to pre-trained weights. Default: None.
        base_momentum (float): The base momentum coefficient for the target network.
            Default: 0.996.
    """

    def __init__(
        self,
        backbone: Dict[str, Any],
        neck: Optional[Dict[str, Any]] = None,
        head: Optional[Dict[str, Any]] = None,
        pretrained: Optional[str] = None,
        base_momentum: float = 0.996,
        **kwargs,  # pylint: disable=unused-argument
    ):

        super().__init__()

        # build backbone
        self.online_backbone = build_backbone(backbone)
        self.target_backbone = build_backbone(backbone)

        # build projector
        self.online_projector = build_neck(neck)
        self.target_projector = build_neck(neck)

        # build head with predictor
        self.head = build_head(head)

        self.init_weights(pretrained=pretrained)

        self.base_momentum = base_momentum
        self.momentum = base_momentum

        # Hooks for super_type transparent weight save
        self._register_state_dict_hook(self.state_dict_hook)

    def init_weights(self, pretrained: Optional[str] = None):
        """Initialize the weights of model.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Default: None.
        """
        if pretrained is not None:
            logger.info(f"load model from: {pretrained}")

        # init backbone
        self.online_backbone.init_weights(pretrained=pretrained)
        for param_ol, param_tgt in zip(self.online_backbone.parameters(), self.target_backbone.parameters()):
            param_tgt.data.copy_(param_ol.data)
            param_tgt.requires_grad = False
            param_ol.requires_grad = True

        # init projector
        self.online_projector.init_weights(init_linear="kaiming")
        for param_ol, param_tgt in zip(self.online_projector.parameters(), self.target_projector.parameters()):
            param_tgt.data.copy_(param_ol.data)
            param_tgt.requires_grad = False
            param_ol.requires_grad = True

        # init the predictor
        self.head.init_weights()

    @torch.no_grad()
    def _momentum_update(self):
        """Momentum update of the target network."""
        for param_ol, param_tgt in zip(self.online_backbone.parameters(), self.target_backbone.parameters()):
            param_tgt.data = param_tgt.data * self.momentum + param_ol.data * (1.0 - self.momentum)

        for param_ol, param_tgt in zip(self.online_projector.parameters(), self.target_projector.parameters()):
            param_tgt.data = param_tgt.data * self.momentum + param_ol.data * (1.0 - self.momentum)

    @torch.no_grad()
    def momentum_update(self):
        """Momentum update of the target network."""
        self._momentum_update()

    def forward(self, img1: torch.Tensor, img2: torch.Tensor, **kwargs):
        """Forward computation during training.

        Args:
            img1 (Tensor): Input of two concatenated images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.
            img2 (Tensor): Input of two concatenated images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        proj_1 = self.online_projector(self.online_backbone(img1))
        proj_2 = self.online_projector(self.online_backbone(img2))
        with torch.no_grad():
            proj_1_tgt = self.target_projector(self.target_backbone(img1)).clone().detach()
            proj_2_tgt = self.target_projector(self.target_backbone(img2)).clone().detach()

        loss = self.head(proj_1, proj_2_tgt)["loss"] + self.head(proj_2, proj_1_tgt)["loss"]
        return dict(loss=loss)

    def train_step(self, data: Dict[str, Any], optimizer):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating are also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.
                ``loss`` is a tensor for back propagation, which can be a
                weighted sum of multiple losses.
                ``log_vars`` contains all the variables to be sent to the
                logger.
                ``num_samples`` indicates the batch size (when the model is
                DDP, it means the batch size on each GPU), which is used for
                averaging the logs.
        """
        losses = self(**data)
        loss, log_vars = self._parse_losses(losses)

        outputs = dict(loss=loss, log_vars=log_vars, num_samples=len(data["img1"].data))

        return outputs

    def val_step(self, *args):
        """Disable validation step during self-supervised learning."""
        pass

    def _parse_losses(self, losses: Dict[str, Any]):
        """Parse loss dictionary.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor
                which may be a weighted sum of all losses, log_vars contains
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            elif isinstance(loss_value, dict):
                for name, value in loss_value.items():
                    log_vars[name] = value
            else:
                raise TypeError(f"{loss_name} is not a tensor or list of tensors")

        loss = sum(_value for _key, _value in log_vars.items() if "loss" in _key)

        log_vars["loss"] = loss
        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars

    @staticmethod
    def state_dict_hook(module, state_dict, prefix, *args, **kwargs):
        """Save only online backbone as output state_dict."""
        logger.info("----------------- BYOL.state_dict_hook() called")
        for k in list(state_dict.keys()):
            v = state_dict.pop(k)
            if not prefix or k.startswith(prefix):
                k = k.replace(prefix, "", 1)
                if k.startswith("online_backbone."):
                    k = k.replace("online_backbone.", "", 1)
                k = prefix + k
            state_dict[k] = v
        return state_dict



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# from datetime import datetime as dt

from .stage import Stage
from .utils.config_utils import copy_config


class Workflow(object):
    def __init__(self, stages, workflow_hooks=None):
        if not isinstance(stages, list):
            raise ValueError("stages parameter should be the list of Stage instance")
        if len(stages) == 0:
            raise ValueError("required one or more stage for the workflow")
        for stage in stages:
            if not isinstance(stage, Stage):
                raise ValueError("stages parameter should be the list of Stage instance")
        if workflow_hooks is not None and not isinstance(workflow_hooks, list):
            raise ValueError("workflow_hooks should be a list")

        self.stages = stages
        self.workflow_hooks = workflow_hooks
        self.results = {}
        self.context = {stage.name: {} for stage in stages}

    def _call_wf_hooks(self, fname, stage_idx=-1):
        if self.workflow_hooks is not None:
            for hook in self.workflow_hooks:
                getattr(hook, fname)(self, stage_idx, self.results)

    def run(self, **kwargs):
        model_cfg = kwargs.get("model_cfg", None)
        data_cfg = kwargs.get("data_cfg", None)
        model_ckpt = kwargs.get("model_ckpt", None)
        # output_path = kwargs.get('output_path', '.')
        mode = kwargs.get("mode", "train")
        ir_model_path = kwargs.get("ir_model_path", None)
        ir_weight_path = kwargs.get("ir_weight_path", None)
        ir_weight_init = kwargs.get("ir_weight_init", False)

        self._call_wf_hooks("before_workflow")
        for i, stage in enumerate(self.stages):
            self._call_wf_hooks("before_stage", i)

            # create keyword arguments that will be passed to stage.run() refer to input map defined in config
            stage_kwargs = dict()
            if hasattr(stage, "input"):
                for arg_name, arg in stage.input.items():
                    stage_name = arg.get("stage_name", None)
                    output_key = arg.get("output_key", None)
                    if stage_name is None or output_key is None:
                        raise ValueError(f"'stage_name' and 'output_key' attributes are required for the '{arg_name}'")
                    stage_kwargs[arg_name] = self.context[stage_name].get(output_key, None)

            # context will keep the results(path to model, etc) of each stage
            # stage.run() returns a dict and each output data will be stored in each output key defined in config
            self.context[stage.name] = stage.run(
                stage_idx=i,
                mode=mode,
                # model_cfg and data_cfg can be changed by each stage. need to pass cloned one for the other stages
                # note that mmcv's Config object manage its attributes inside of _cfg_dict so need to copy it as well
                model_cfg=copy_config(model_cfg) if model_cfg is not None else model_cfg,
                data_cfg=copy_config(data_cfg) if data_cfg is not None else data_cfg,
                model_ckpt=model_ckpt,
                # output_path=output_path+'/stage{:02d}_{}'.format(i, stage.name),
                ir_model_path=ir_model_path,
                ir_weight_path=ir_weight_path,
                ir_weight_init=ir_weight_init,
                **stage_kwargs,
            )
            # TODO: save context as pickle after each stage??
            self._call_wf_hooks("after_stage", i)
        self._call_wf_hooks("after_workflow")
